import json
import time
import typing as t
from datetime import datetime
from io import BytesIO
from logging import LoggerAdapter

import requests
from pydantic import Field

from hrflow_connectors.core import (
    DataType,
    FieldType,
    ParametersModel,
    Warehouse,
    WarehouseReadAction,
    WarehouseWriteAction,
)
from hrflow_connectors.core.warehouse import ReadMode
from hrflow_connectors.v1.connectors.bullhorn.schemas import (
    BullhornJob,
    BullhornProfile,
)
from hrflow_connectors.v1.connectors.bullhorn.utils.authentication import auth


class BaseParameters(ParametersModel):
    client_id: str = Field(
        ...,
        description="Client identifier for Bullhorn",
        repr=False,
        field_type=FieldType.Auth,
    )
    client_secret: str = Field(
        ...,
        description="Client secret identifier for Bullhorn",
        repr=False,
        field_type=FieldType.Auth,
    )
    password: str = Field(
        ...,
        description="Password for Bullhorn login",
        repr=False,
        field_type=FieldType.Auth,
    )
    username: str = Field(
        ...,
        description="Username for Bullhorn login",
        repr=False,
        field_type=FieldType.Auth,
    )


class WriteProfilesParameters(BaseParameters):
    pass


class WriteApplicationsParameters(BaseParameters):
    job_id: str = Field(
        ...,
        description="id for the job in Bullhorn",
        repr=False,
        field_type=FieldType.Auth,
    )

    candidate_id: str = Field(
        None,
        description="id of the profile in Bullhorn",
        repr=False,
        field_type=FieldType.Auth,
    )

    status_when_created: str = Field(
        ...,
        description="The status of the application when created in Bullhorn",
        repr=False,
        field_type=FieldType.Auth,
    )

    source: str = Field(
        None,
        description="The source of the application to be created in Bullhorn",
        repr=False,
        field_type=FieldType.Auth,
    )


class ReadParameters(BaseParameters):
    last_modified_date: str = Field(
        ...,
        description="Last Modified Date in timestamp",
        repr=False,
        field_type=FieldType.Auth,
    )

    fields: str = Field(
        ...,
        description="Fields to be retrieved from Bullhorn",
        repr=False,
        field_type=FieldType.Auth,
    )

    query: str = Field(
        ...,
        description="the query parameters",
        repr=False,
        field_type=FieldType.Auth,
    )

    count: int = Field(
        ...,
        description="Number of items to be returned",
        repr=False,
        field_type=FieldType.Auth,
    )


class ReadProfileParameters(BaseParameters):
    pass


def write(
    adapter: LoggerAdapter,
    parameters: WriteProfilesParameters,
    profiles: t.Iterable[t.Dict],
) -> t.List[t.Dict]:
    adapter.info("Pushing {} profiles".format(len(profiles)))
    failed_profiles = []
    authentication = auth(
        parameters.username,
        parameters.password,
        parameters.client_id,
        parameters.client_secret,
    )

    for profile in profiles:
        # Split profile in four parts: Body, Education, Experience and Attachements
        profile_body_dict = profile
        create_profile_body = profile_body_dict["create_profile_body"]
        enrich_profile_education = profile_body_dict["enrich_profile_education"]
        enrich_profile_experience = profile_body_dict["enrich_profile_experience"]
        enrich_profile_attachment = profile_body_dict["enrich_profile_attachment"]

        rest_url = authentication["restUrl"]
        params = {"BhRestToken": authentication["BhRestToken"]}

        candidate_url = rest_url + "entity/Candidate"
        response = requests.put(
            url=candidate_url, json=create_profile_body, params=params
        )

        # Unable to push profile
        if response.status_code // 100 != 2:
            adapter.error(
                "Failed to push profile from to Bullhorn status_code={} response={}"
                .format(response.status_code, response.text)
            )
            failed_profiles.append(profile)
            continue

        candidate_id = response.json()
        candidate_id = str(candidate_id["changedEntityId"])

        # Enrich profile education
        for education in enrich_profile_education:
            education["candidate"]["id"] = candidate_id
            education_url = rest_url + "entity/CandidateEducation"
            response = requests.put(url=education_url, json=education, params=params)

        # Enrich profile experience
        for experience in enrich_profile_experience:
            experience["candidate"]["id"] = candidate_id
            experience_url = rest_url + "entity/CandidateWorkHistory"
            response = requests.put(url=experience_url, json=experience, params=params)

        # Enrich profile attachements
        for attachment in enrich_profile_attachment:
            attachment_url = rest_url + "file/Candidate/" + str(candidate_id)
            response = requests.put(url=attachment_url, json=attachment, params=params)

    return failed_profiles


def authenticate(parameters):
    return auth(
        parameters.username,
        parameters.password,
        parameters.client_id,
        parameters.client_secret,
    )


def make_request(method, url, params, adapter, json=None):
    response = method(url, params=params, data=json)
    if response.status_code == 401:
        adapter.info("Auth token expired, regenerating...")
        auth_info = authenticate()
        params["BhRestToken"] = auth_info["BhRestToken"]
        response = method(url, params=params, data=json)
    return response


def handle_response(response, adapter):
    if not response.ok:
        adapter.error(
            f"Request failed, status_code={response.status_code},"
            f" response={response.text}"
        )
        return None
    return response.json()


def search_entity(entity, rest_url, bh_rest_token, query, fields, adapter):
    search_url = f"{rest_url}search/{entity}"
    params = {
        "BhRestToken": bh_rest_token,
        "query": query,
        "fields": fields,
        "sort": "id",
    }
    response = make_request(requests.get, search_url, params, adapter)
    return handle_response(response, adapter)


def create_or_update_entity(entity, rest_url, params, data, adapter, entity_id=None):
    url = f"{rest_url}entity/{entity}"
    method = requests.post if entity_id else requests.put
    if entity_id:
        url = f"{url}/{entity_id}"
    response = make_request(method, url, params, adapter, json.dumps(data))
    return handle_response(response, adapter)


def check_entity_files(entity, rest_url, params, entity_id, adapter):
    url = f"{rest_url}entityFiles/{entity}/{entity_id}"
    response = requests.get(url, params=params)
    return handle_response(response, adapter)


def write_application(
    adapter: LoggerAdapter,
    parameters: WriteApplicationsParameters,
    profiles: t.Iterable[t.Dict],
) -> t.List[t.Dict]:
    failed_profiles = []
    auth_info = authenticate(parameters)
    rest_url = auth_info["restUrl"]
    bh_rest_token = auth_info["BhRestToken"]
    params = {"BhRestToken": bh_rest_token}
    adapter.info(f"connexion info {params}, rest_url: {rest_url}")

    for profile in profiles:
        comment = profile.pop("comment", None)
        if parameters.candidate_id:
            candidate_id = parameters.candidate_id
        if not parameters.candidate_id:
            attachment = profile.pop("attachment")
            profile["source"] = parameters.source or profile.get("source")
            profile["status"] = parameters.status_when_created or profile.get("status")
            email = profile["email"]
            adapter.info(f"checking if candidate with {email} already exists")
            search_results = search_entity(
                "Candidate",
                rest_url,
                bh_rest_token,
                f"(email:{email} OR email2:{email}) AND isDeleted:0",
                (
                    "id,isDeleted,dateAdded,status,source,email,"
                    "firstName,lastName,name,mobile,address"
                ),
                adapter,
            )

            if not search_results:
                failed_profiles.append(profile)
                continue
            adapter.info(f"search profile response {search_results}")
            candidate_exists = search_results["count"] > 0
            candidate_data = search_results["data"][0] if candidate_exists else {}
            candidate_id = candidate_data.get("id") if candidate_exists else None

            if candidate_exists:
                profile.update(
                    {
                        "firstName": candidate_data.get("firstName") or profile.get(
                            "firstName"
                        ),
                        "lastName": candidate_data.get("lastName") or profile.get(
                            "lastName"
                        ),
                        "name": candidate_data.get("name") or profile.get("name"),
                        "address": candidate_data.get("address") or profile.get(
                            "address"
                        ),
                        "mobile": candidate_data.get("mobile") or profile.get("mobile"),
                        "status": candidate_data.get("status") or profile.get("status"),
                    }
                )
            adapter.info("creating or updating the candidate")
            candidate_response = create_or_update_entity(
                "Candidate", rest_url, params, profile, adapter, candidate_id
            )
            if not candidate_response:
                failed_profiles.append(profile)
                continue
            adapter.info(f"candidate creation response {candidate_response}")
            if not candidate_exists:
                candidate_id = candidate_response.get("changedEntityId")

            attachment_exists = False
            if candidate_exists and attachment:
                entity_files = check_entity_files(
                    "Candidate", rest_url, params, candidate_id, adapter
                )
                if entity_files:
                    attachments = entity_files.get("EntityFiles", [])
                    if attachments and attachment["name"] == attachments[0]["name"]:
                        attachment_exists = True
            adapter.info(f"attachment for the candidate exists {attachment_exists}")
            if not attachment_exists and attachment:
                attachment_response = make_request(
                    requests.put,
                    f"{rest_url}file/Candidate/{candidate_id}",
                    params,
                    adapter,
                    json.dumps(attachment),
                )
                if not handle_response(attachment_response, adapter):
                    failed_profiles.append(profile)
                    continue
                attachment_response = handle_response(attachment_response, adapter)
                adapter.info(f"attachment response {attachment_response}")
            adapter.info(
                "Verifying if candidate had already applied for the job"
                f" {parameters.job_id}"
            )
        job_submission_results = search_entity(
            "JobSubmission",
            rest_url,
            bh_rest_token,
            f"candidate.id:{candidate_id} AND jobOrder.id:{parameters.job_id}",
            "id,status,dateAdded",
            adapter,
        )

        if not job_submission_results:
            failed_profiles.append(profile)
            continue
        adapter.info(f"search job_submission response {job_submission_results}")
        job_submission_exists = job_submission_results["count"] > 0
        job_submission_id = (
            job_submission_results["data"][0]["id"] if job_submission_exists else None
        )

        job_submission_payload = {
            "candidate": {"id": candidate_id},
            "jobOrder": {"id": parameters.job_id},
            "status": parameters.status_when_created,
            "source": parameters.source,
            "dateWebResponse": int(time.time() * 1000),
        }

        if comment:
            job_submission_payload["comments"] = comment

        adapter.info("Creating or updating if candidate jobSubmission")
        job_submission_response = create_or_update_entity(
            "JobSubmission",
            rest_url,
            params,
            job_submission_payload,
            adapter,
            job_submission_id,
        )
        if not job_submission_response:
            failed_profiles.append(profile)
        adapter.info(f"creation of job_submission response {job_submission_response}")
    return failed_profiles


def read_jobs(
    adapter: LoggerAdapter,
    parameters: ReadParameters,
    read_mode: t.Optional[ReadMode] = None,
    read_from: t.Optional[str] = None,
) -> t.Iterable[t.Dict]:
    start = 0
    auth_retries = 0
    total_returned = 0
    should_break = False
    authentication = auth(
        parameters.username,
        parameters.password,
        parameters.client_id,
        parameters.client_secret,
    )
    if read_mode is ReadMode.sync:
        if parameters.last_modified_date is None:
            raise Exception("last_modified_date cannot be None in ReadMode.sync")
        last_modified_date = parameters.last_modified_date
    else:
        if parameters.last_modified_date is not None:
            adapter.warning(
                "last_modified_date is ignored in ReadMode.incremental, using"
                " read_from instead"
            )
        if read_from:
            try:
                read_from = json.loads(read_from)
                last_modified_date = read_from["last_modified_date"]
                last_id = read_from["last_id"]
            except json.JSONDecodeError as e:
                raise Exception(f"Failed to JSON parse read_from={read_from} error={e}")
            except KeyError as e:
                raise Exception(
                    "Failed to find expected key in"
                    f" read_from={read_from} error={repr(e)}"
                )
        else:
            last_modified_date = parameters.last_modified_date

    last_modified_date_filter = transform_timestamp(last_modified_date)
    if not last_modified_date_filter:
        raise Exception(
            "error while applying a transformation date on last modified date to"
            " perform filtering"
        )

    while True:
        try:
            query = (
                f"{parameters.query} AND "
                f"dateLastModified:[{last_modified_date_filter} TO *]"
            )
            jobs_url = f"{authentication['restUrl']}search/JobOrder"
            params = {
                "query": query,
                "fields": parameters.fields,
                "sort": "dateLastModified,id",
                "start": str(start),
            }

            if parameters.count:
                params["count"] = parameters.count

            headers = {"BhRestToken": authentication["BhRestToken"]}

            response = requests.get(url=jobs_url, params=params, headers=headers)

            response = response.json()
            start = response["start"] + response["count"]
            data = response["data"]
            for job in data:
                if parameters.count and total_returned >= parameters.count:
                    should_break = True
                    break
                if (
                    read_mode is ReadMode.incremental
                    and job["dateLastModified"] == last_modified_date
                    and job["id"] <= last_id
                ):
                    adapter.info("job with id <= last_id")
                    continue
                yield job
                total_returned += 1
            if should_break:
                break
            if start >= response["total"]:
                break

        except requests.HTTPError as e:
            if e.response.status_code == 401:
                adapter.info(
                    "Received 401 error. Retrying authentication to continue fetching"
                    " jobs."
                )
                if auth_retries > 2:
                    raise Exception(
                        f" retries the authentication {auth_retries}"
                        " will stop the execution"
                    )

                authentication = auth(
                    parameters.username,
                    parameters.password,
                    parameters.client_id,
                    parameters.client_secret,
                    refresh_token=authentication["refresh_token"],
                )
                auth_retries += 1
                continue
            else:
                adapter.error("Failed to fetch jobs from Bullhorn.")
                raise e


def read_profiles_parsing(
    adapter: LoggerAdapter,
    parameters: ReadParameters,
    read_mode: t.Optional[ReadMode] = None,
    read_from: t.Optional[str] = None,
) -> t.Iterable[t.Dict]:
    authentication = auth(
        parameters.username,
        parameters.password,
        parameters.client_id,
        parameters.client_secret,
    )

    start = 0
    auth_retries = 0
    total_returned = 0
    should_break = False
    last_id = None

    if read_mode is ReadMode.sync:
        if parameters.last_modified_date is None:
            raise Exception("last_modified_date cannot be None in ReadMode.sync")
        last_modified_date = parameters.last_modified_date
    else:
        if parameters.last_modified_date is not None:
            adapter.warning(
                "last_modified_date is ignored in ReadMode.incremental, using"
                " read_from instead"
            )
        if read_from:
            try:
                read_from = json.loads(read_from)
                last_modified_date = read_from["last_modified_date"]
                last_id = read_from["last_id"]
            except json.JSONDecodeError as e:
                raise Exception(f"Failed to JSON parse read_from={read_from} error={e}")
            except KeyError as e:
                raise Exception(
                    "Failed to find expected key in"
                    f" read_from={read_from} error={repr(e)}"
                )
        else:
            last_modified_date = parameters.last_modified_date

    last_modified_date_filter = transform_timestamp(last_modified_date)
    if not last_modified_date_filter:
        raise Exception(
            "Error while applying a transformation date on last modified date to"
            " perform filtering"
        )

    while True:
        try:
            query = (
                f"{parameters.query} AND "
                f"dateLastModified:[{last_modified_date_filter} TO *]"
            )

            profiles_url = f"{authentication['restUrl']}search/Candidate"
            params = {
                "query": query,
                "fields": parameters.fields,
                "sort": "dateLastModified,id",
                "start": str(start),
            }

            if parameters.count:
                params["count"] = parameters.count

            headers = {"BhRestToken": authentication["BhRestToken"]}

            response = requests.get(url=profiles_url, params=params, headers=headers)
            if response.status_code // 100 != 2:
                adapter.error(
                    "Failed to pull profiles from Bullhorn"
                    f" status_code={response.status_code} response={response.text}"
                )
                raise Exception("Failed to pull profiles from Bullhorn")
            response = response.json()

            start = response["start"] + response["count"]
            data = response["data"]
            total = response["total"]

            for profile in data:
                if parameters.count and total_returned >= parameters.count:
                    should_break = True
                    break
                if (
                    read_mode is ReadMode.incremental
                    and profile["dateLastModified"] == last_modified_date
                    and profile["id"] <= last_id
                ):
                    adapter.info("Profile with id <= last_id")
                    continue

                profile["cvFile"] = None
                url_files = (
                    authentication["restUrl"]
                    + "entityFiles/Candidate/"
                    + str(profile["id"])
                )
                headers = {"BhRestToken": authentication["BhRestToken"]}
                response = requests.get(url=url_files, headers=headers)
                response = response.json()

                last_cv = None
                curr_entity = None
                if len(response["EntityFiles"]) > 0:
                    for entity_file in response["EntityFiles"]:
                        if entity_file["type"] == "Resume":
                            if not curr_entity:
                                curr_entity = entity_file
                                last_cv = entity_file["id"]
                            elif curr_entity["dateAdded"] < entity_file["dateAdded"]:
                                curr_entity = entity_file
                                last_cv = entity_file["id"]

                if last_cv is not None:
                    url_cv = (
                        authentication["restUrl"]
                        + "/file/Candidate/"
                        + str(profile["id"])
                        + "/"
                        + str(last_cv)
                        + "/raw"
                    )
                    response = requests.get(url=url_cv, headers=headers)

                    file = response.content
                    profile_file = BytesIO(file)
                    profile["cvFile"] = profile_file

                total_returned += 1
                yield profile

            if should_break:
                break

            if start >= total:
                break

        except requests.HTTPError as e:
            if e.response.status_code == 401:
                adapter.info(
                    "Received 401 error. Retrying authentication to continue fetching"
                    " profiles."
                )
                if auth_retries > 2:
                    raise Exception(
                        f"Retries exceeded for authentication ({auth_retries})."
                        " Stopping execution."
                    )

                authentication = auth(
                    parameters.username,
                    parameters.password,
                    parameters.client_id,
                    parameters.client_secret,
                    refresh_token=authentication["refresh_token"],
                )
                auth_retries += 1
                continue
            else:
                adapter.error("Failed to fetch profiles from Bullhorn.")
                raise e


def read_profiles(
    adapter: LoggerAdapter,
    parameters: ReadProfileParameters,
    read_mode: t.Optional[ReadMode] = None,
    read_from: t.Optional[str] = None,
) -> t.Iterable[t.Dict]:
    authentication = auth(
        parameters.username,
        parameters.password,
        parameters.client_id,
        parameters.client_secret,
    )
    fields = """firstName,lastName,name,email,mobile,dateOfBirth,gender,address,dateAvailable,status,employeeType,activePlacements,
    skillSet,id,educations,workHistories"""
    start = 0
    while True:
        profiles_url = (
            authentication["restUrl"]
            + "myCandidates?fields="
            + fields
            + "&start="
            + str(start)
        )
        headers = {"BhRestToken": authentication["BhRestToken"]}

        response = requests.get(url=profiles_url, headers=headers)
        if response.status_code // 100 != 2:
            adapter.error(
                "Failed to pull profiles from Bullhorn status_code={} response={}"
                .format(response.status_code, response.text)
            )
            raise Exception("Failed to pull profiles from Bullhorn")
        response = response.json()

        start = response["start"] + response["count"]
        total = response["total"]
        data = response["data"]

        for profile in data:
            # Enrich profile education
            education_ids = []
            educations = []
            for ed in profile["educations"]["data"]:
                education_ids.append(ed["id"])
            for id in education_ids:
                education_url = (
                    authentication["restUrl"]
                    + "entity/CandidateEducation/"
                    + str(id)
                    + "?fields=*"
                )
                response = requests.get(url=education_url, headers=headers)
                response = response.json()
                educations.append(response["data"])

            # Enrich profile work history
            work_history_ids = []
            work_histories = []
            for work_history in profile["workHistories"]["data"]:
                work_history_ids.append(work_history["id"])
            for id in work_history_ids:
                work_history_url = (
                    authentication["restUrl"]
                    + "entity/CandidateWorkHistory/"
                    + str(id)
                    + "?fields=*"
                )
                response = requests.get(url=work_history_url, headers=headers)
                response = response.json()
                work_histories.append(response["data"])

            profile["educations"] = educations
            profile["workHistories"] = work_histories

            yield profile

        if start >= total:
            break


def transform_timestamp(timestamp: t.Optional[t.Union[float, int]]) -> t.Optional[str]:
    if timestamp is None:
        return None
    # Convert the Unix timestamp (in milliseconds) to a datetime object
    dt = datetime.utcfromtimestamp(int(timestamp) / 1000)
    # Format the datetime object to something like 20221215121030
    transformed_date = dt.strftime("%Y%m%d%H%M%S")
    return transformed_date


def item_to_read_from(item: t.Dict) -> str:
    return json.dumps(
        dict(last_modified_date=item["dateLastModified"], last_id=item["id"])
    )


BullhornProfileWarehouse = Warehouse(
    name="Bullhorn Profiles",
    data_schema=BullhornProfile,
    data_type=DataType.profile,
    write=WarehouseWriteAction(
        parameters=WriteProfilesParameters, function=write, endpoints=[]
    ),
    read=WarehouseReadAction(
        parameters=ReadProfileParameters, function=read_profiles, endpoints=[]
    ),
)

BullhornApplicationWarehouse = Warehouse(
    name="Bullhorn Applications",
    data_schema=BullhornProfile,
    data_type=DataType.profile,
    write=WarehouseWriteAction(
        parameters=WriteApplicationsParameters, function=write_application, endpoints=[]
    ),
)

BullhornProfileParsingWarehouse = Warehouse(
    name="Bullhorn Profiles",
    data_schema=BullhornProfile,
    data_type=DataType.profile,
    read=WarehouseReadAction(
        parameters=ReadParameters,
        function=read_profiles_parsing,
        endpoints=[],
        supports_incremental=True,
        item_to_read_from=item_to_read_from,
    ),
)

BullhornJobWarehouse = Warehouse(
    name="Bullhorn Jobs",
    data_schema=BullhornJob,
    data_type=DataType.job,
    read=WarehouseReadAction(
        parameters=ReadParameters,
        function=read_jobs,
        supports_incremental=True,
        item_to_read_from=item_to_read_from,
    ),
)



import typing as t
from logging import LoggerAdapter

import mysql.connector
from mysql.connector import Error
from mysql.connector.connection import MySQLConnection


def connect_to_database(
    host: str, port: int, user: str, password: str, database: str
) -> t.Optional[MySQLConnection]:
    """Establish a connection to the database.

    :param host: The hostname of the database server.
    :param port: The port number to connect to.
    :param user: The username to authenticate with.
    :param password: The password to authenticate with.
    :param database: The name of the database to connect to.
    :return: A MySQLConnection object if the connection was successful, otherwise None.
    """
    try:
        connection = mysql.connector.connect(
            host=host, port=port, user=user, password=password, database=database
        )
        if isinstance(connection, MySQLConnection) and connection.is_connected():
            print("Connected to the database")
            return connection
    except Error as e:
        print(f"Error: {e}")
    return None


# Function to query the database and return results as a dictionary
def query_database(
    connection: MySQLConnection, query: str, params: t.Optional[t.Tuple] = None
) -> t.Optional[t.List[t.Dict]]:
    """
    Execute a query with optional parameters and return results as a dictionary.

    :param connection: Database connection object.
    :param query: SQL query string.
    :param params: Tuple or list of parameters to use in the query.
    :return: List of dictionaries containing the query results.
    """
    cursor = connection.cursor(dictionary=True)
    try:
        if params:
            cursor.execute(query, params)
        else:
            cursor.execute(query)
        results = cursor.fetchall()
    except Exception as e:
        print(f"Error executing query: {e}")
        results = []
    finally:
        cursor.close()
    return results


# Function to insert a new object into a table
def insert_object(
    adapter: LoggerAdapter,
    connection: MySQLConnection,
    table_name: str,
    object_data: t.Dict,
) -> t.Optional[int]:
    """
    Inserts a new object into the database.

    :param adapter: LoggerAdapter for logging.
    :param connection: MySQL database connection.
    :param table_name: The name of the table to insert into.
    :param object_data: Dictionary of column-value pairs to insert.
    :return: The ID of the newly inserted row, or None if the operation failed.
    """

    placeholders = ", ".join(["%s"] * len(object_data))
    columns = ", ".join(
        [f"`{col}`" for col in object_data.keys()]
    )  # Enclose column names in backticks
    sql = f"INSERT INTO {table_name} ({columns}) VALUES ({placeholders})"
    cursor = connection.cursor()
    try:
        cursor.execute(sql, list(object_data.values()))
        connection.commit()
        inserted_id = cursor.lastrowid  # Get the ID of the newly inserted row
        adapter.info("Object inserted successfully with ID:", inserted_id)
        return inserted_id  # Return the ID to the caller
    except Error as e:
        connection.rollback()
        adapter.error(f"Failed to insert object into the table {table_name}: {e}")
        return None
    finally:
        cursor.close()


def update_object(
    adapter: LoggerAdapter,
    connection: MySQLConnection,
    table_name: str,
    object_data: t.Dict,
    where_clause: t.Dict,
) -> bool:
    """
    Updates an object in the database.

    :param adapter: LoggerAdapter for logging.
    :param connection: MySQL database connection.
    :param table_name: The name of the table to update.
    :param object_data: Dictionary of column-value pairs to update.
    :param where_clause: Dictionary of conditions to identify the rows to update.
    :return: True if the operation was successful, False otherwise.
    """
    set_clause = ", ".join([f"`{col}` = %s" for col in object_data.keys()])
    where_conditions = " AND ".join([f"`{col}` = %s" for col in where_clause.keys()])
    sql = f"UPDATE {table_name} SET {set_clause} WHERE {where_conditions}"

    cursor = connection.cursor()
    try:
        cursor.execute(sql, list(object_data.values()) + list(where_clause.values()))
        connection.commit()
        rows_affected = cursor.rowcount
        adapter.info(f"{rows_affected} rows updated in table {table_name}.")
        return True
    except Error as e:
        connection.rollback()
        adapter.error(f"Failed to update object in the table {table_name}: {e}")
        return False
    finally:
        cursor.close()


# def update_object(
#     adapter: LoggerAdapter,
#     connection: MySQLConnection,
#     table_name: str,
#     object_data: t.Dict,
#     where_clause: t.Dict,
# ) -> bool:
#     set_clause = ", ".join(
#         [
#             f"`{col}` = %s" if object_data[col] is not None else f"`{col}` = NULL"
#             for col in object_data
#         ]
#     )
#     where_conditions = " AND ".join(
#         [
#             f"`{col}` = %s" if where_clause[col] is not None else f"`{col}` IS NULL"
#             for col in where_clause
#         ]
#     )
#     sql = f"UPDATE {table_name} SET {set_clause} WHERE {where_conditions}"

#     cursor = connection.cursor()
#     try:
#         values = [value for value in object_data.values() if value is not None] + [
#             value for value in where_clause.values() if value is not None
#         ]
#         adapter.debug("SQL Query:", sql)
#         adapter.debug("Values:", values)
#         cursor.execute(sql, values)
#         connection.commit()
#         rows_affected = cursor.rowcount
#         if rows_affected == 0:
#             adapter.warning(f"No rows updated in table {table_name}.")
#         else:
#             adapter.info(f"{rows_affected} rows updated in table {table_name}.")
#         return rows_affected > 0
#     except Error as e:
#         connection.rollback()
#         adapter.error(f"Failed to update object in the table {table_name}: {e}")
#         return False
#     finally:
#         cursor.close()



from hrflow_connectors.v2.connectors.workable.aisles import (
    AuthParameters,
    JobsAisle,
    ProfilesAisle,
)
from hrflow_connectors.v2.core.warehouse import Warehouse

WorkableWarehouse = Warehouse(
    auth=AuthParameters,
    aisles=(
        JobsAisle,
        ProfilesAisle,
    ),
)



import enum
import typing as t
from logging import LoggerAdapter

import requests
from msgspec import Meta, Struct
from typing_extensions import Annotated

from hrflow_connectors.v2.connectors.smartrecruiters.schemas import (
    SmartRecruitersJob,
    SmartRecruitersProfile,
)
from hrflow_connectors.v2.core.common import Entity
from hrflow_connectors.v2.core.warehouse import (
    Aisle,
    Criterias,
    Endpoint,
    Endpoints,
    ReadOperation,
    WriteOperation,
    merge,
)

SMARTRECRUITERS_JOBS_ENDPOINT = "https://api.smartrecruiters.com/jobs"
SMARTRECRUITERS_PROFILES_ENDPOINT = "https://api.smartrecruiters.com/candidates"
SMARTRECRUITERS_ENDPOINT_LIMIT = 100

GET_ALL_JOBS_ENDPOINT = Endpoint(
    name="Get all jobs",
    description=(
        "Endpoint to search jobs by traditional params (offset, limit...)"
        " and get the list of all jobs with their ids, the request method"
        " is `GET`"
    ),
    url="https://developers.smartrecruiters.com/reference/jobsall-1",
)
GET_JOB_ENDPOINT = Endpoint(
    name="Get job",
    description=(
        "Endpoint to get the content of a job with a given id, a jobId parameter is"
        " required, the request method is `GET`"
    ),
    url="https://developers.smartrecruiters.com/reference/jobsget-1",
)
UPDATE_JOB_ENDPOINT = Endpoint(
    name="Update job",
    description=(
        "Endpoint to update a job with a given id, a jobId parameter is required, the"
        " request method is `PUT`"
    ),
    url="https://developers.smartrecruiters.com/reference/jobspatch-1",
)
GET_CANDIDATE_ENDPOINT = Endpoint(
    name="Get candidate",
    description=(
        "Endpoint to get the content of a candidate with a given id, a candidateId"
        " parameter is required, the request method is `GET`"
    ),
    url="https://developers.smartrecruiters.com/reference/candidatesget-1",
)

POST_CANDIDATE_ENDPOINT = Endpoint(
    name="Post Candidate",
    description=(
        "Endpoint to create a new candidate and assign to a talent pool, the request"
        " method is `POST`"
    ),
    url="https://developers.smartrecruiters.com/reference/candidatesadd-1",
)

UPDATE_CANDIDATE_ENDPOINT = Endpoint(
    name="Update Candidate",
    description=(
        "Endpoint to update a candidate with a given id, a candidateId parameter is"
        " required, the request method is `PUT`"
    ),
    url="https://developers.smartrecruiters.com/reference/candidatespatch-1",
)

DELETE_CANDIDATE_ENDPOINT = Endpoint(
    name="Delete Candidate",
    description=(
        "Endpoint to delete a candidate with a given id, a candidateId parameter is"
        " required, the request method is `DELETE`"
    ),
    url="https://developers.smartrecruiters.com/reference/candidatesdelete-1",
)


class JobPostingStatus(str, enum.Enum):
    public = "PUBLIC"
    internal = "INTERNAL"
    not_published = "NOT_PUBLISHED"
    private = "PRIVATE"


class JobStatus(str, enum.Enum):
    created = "CREATED"
    sourcing = "SOURCING"
    filled = "FILLED"
    interview = "INTERVIEW"
    offer = "OFFER"
    cancelled = "CANCELLED"
    on_hold = "ON_HOLD"


class OnboardingStatus(str, enum.Enum):
    ready_to_onboard = "READY_TO_ONBOARD"
    onboarding_successful = "ONBOARDING_SUCCESSFUL"
    onboarding_failed = "ONBOARDING_FAILED"


class AuthParameters(Struct):
    x_smart_token: Annotated[
        str,
        Meta(
            description="X-SmartToken used to access SmartRecruiters API",
        ),
    ]


class ReadJobsParameters(Struct):
    query: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "Case insensitive full-text query against job title e.g. java developer"
            ),
        ),
    ] = None
    updated_after: Annotated[
        t.Optional[str],
        Meta(
            description="ISO8601-formatted time boundaries for the job update time",
        ),
    ] = None
    posting_status: Annotated[
        t.Optional[JobPostingStatus],
        Meta(
            description="Posting status of a job. One of {}".format(
                [e.value for e in JobPostingStatus]
            ),
        ),
    ] = None
    job_status: Annotated[
        t.Optional[JobStatus],
        Meta(
            description="Status of a job. One of {}".format(
                [e.value for e in JobStatus]
            ),
        ),
    ] = None
    limit: Annotated[
        int,
        Meta(description="Number of items to pull from SmartRecruiters at a time."),
    ] = SMARTRECRUITERS_ENDPOINT_LIMIT


class ReadProfilesParameters(Struct):
    query: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "keyword search, for more infromation see SmartRecruiters HelpCenter"
            ),
        ),
    ] = None
    job_ids: Annotated[
        t.Optional[t.List[str]],
        Meta(
            description="List of job ids to filter candidates by",
        ),
    ] = None
    tags: Annotated[
        t.Optional[t.List[str]],
        Meta(
            description="List of tags to filter candidates by",
        ),
    ] = None
    onboarding_status: Annotated[
        t.Optional[OnboardingStatus],
        Meta(
            description="Onboarding status of a candidate",
        ),
    ] = None
    status: Annotated[
        t.Optional[t.List[str]],
        Meta(
            description="candidate’s status filter in a context of a job",
        ),
    ] = None
    limit: Annotated[
        int,
        Meta(
            description=(
                "Number of items to pull from SmartRecruiters at a time. Not matter"
                " what value is supplied it is capped at {}".format(
                    SMARTRECRUITERS_ENDPOINT_LIMIT
                )
            ),
        ),
    ] = SMARTRECRUITERS_ENDPOINT_LIMIT
    updated_after: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "ISO8601-formatted time boundaries for the candidate update time"
            ),
        ),
    ] = None


class WriteProfilesParameters(Struct):
    pass


def read_jobs(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
    parameters: ReadJobsParameters,
    incremental: bool,
    incremental_token: t.Optional[str],
) -> t.Iterable[t.Dict]:
    # The limit parameter in the endpoint specifies the number of items per page,
    # not the total number of items returned. Therefore, we need to manually limit
    # the number of items retrieved.

    page = None
    jobs = []
    while True:
        params = dict(
            q=parameters.query,
            updatedAfter=parameters.updated_after,
            postingStatus=parameters.posting_status,
            status=parameters.job_status,
            limit=(
                parameters.limit
                if parameters.limit < SMARTRECRUITERS_ENDPOINT_LIMIT
                else SMARTRECRUITERS_ENDPOINT_LIMIT
            ),
            pageId=page,
        )
        response = requests.get(
            SMARTRECRUITERS_JOBS_ENDPOINT,
            headers={"X-SmartToken": auth_parameters.x_smart_token},
            params=params,
        )
        if response.status_code // 100 != 2:
            adapter.error(
                "Failed to pull jobs from SmartRecruiters params={}"
                " status_code={} response={}".format(
                    params, response.status_code, response.text
                )
            )
            raise Exception("Failed to pull jobs from SmartRecruiters")
        response = response.json()
        jobs += response["content"]
        if len(jobs) == 0 or len(jobs) >= parameters.limit:
            break

        page = response["nextPageId"]

    if len(jobs) > parameters.limit:
        jobs = jobs[: parameters.limit]

    for job in jobs:
        full_job_response = requests.get(
            "{}/{}".format(SMARTRECRUITERS_JOBS_ENDPOINT, job["id"]),
            headers={"X-SmartToken": auth_parameters.x_smart_token},
        )
        if full_job_response.status_code // 100 != 2:
            adapter.error(
                "Failed to pull jobs details from SmartRecruiters job_id={}"
                " status_code={} response={}".format(
                    job["id"],
                    full_job_response.status_code,
                    full_job_response.text,
                )
            )
            raise Exception("Failed to pull jobs from SmartRecruiters")
        yield full_job_response.json()


def read_profiles(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
    parameters: ReadProfilesParameters,
    incremental: bool,
    incremental_token: t.Optional[str],
):
    # The limit parameter in the endpoint specifies the number of items per page,
    # not the total number of items returned. Therefore, we need to manually limit
    # the number of items retrieved.

    page = None
    profiles = []
    while True:
        params = dict(
            q=parameters.query,
            updatedAfter=parameters.updated_after,
            jobId=parameters.job_ids,
            tag=parameters.tags,
            onboardingStatus=parameters.onboarding_status,
            status=parameters.status,
            limit=(
                parameters.limit
                if parameters.limit < SMARTRECRUITERS_ENDPOINT_LIMIT
                else SMARTRECRUITERS_ENDPOINT_LIMIT
            ),
            pageId=page,
        )
        response = requests.get(
            SMARTRECRUITERS_PROFILES_ENDPOINT,
            headers={"x-smarttoken": auth_parameters.x_smart_token},
            params=params,
        )
        if response.status_code // 100 != 2:
            adapter.error(
                "Failed to pull profiles from SmartRecruiters params={}"
                " status_code={} response={}".format(
                    params, response.status_code, response.text
                )
            )
            raise Exception("Failed to pull profiles from SmartRecruiters")
        response = response.json()
        profiles += response["content"]
        if len(profiles) == 0 or len(profiles) >= parameters.limit:
            break

        page = response["nextPageId"]

    if len(profiles) > parameters.limit:
        profiles = profiles[: parameters.limit]

    for profile in profiles:
        full_profile_response = requests.get(
            "{}/{}".format(SMARTRECRUITERS_PROFILES_ENDPOINT, profile["id"]),
            headers={"x-smarttoken": auth_parameters.x_smart_token},
        )
        if full_profile_response.status_code // 100 != 2:
            adapter.error(
                "Failed to pull profile details from SmartRecruiters profile_id={}"
                " status_code={} response={}".format(
                    profile["id"],
                    full_profile_response.status_code,
                    full_profile_response.text,
                )
            )
            raise Exception("Failed to pull profiles from SmartRecruiters")
        yield full_profile_response.json()


def write_profiles(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
    parameters: WriteProfilesParameters,
    items: t.Iterable[t.Dict],
) -> t.List[t.Dict]:
    failed_profiles = []

    for profile in items:
        response = requests.post(
            SMARTRECRUITERS_PROFILES_ENDPOINT,
            headers={"x-smarttoken": auth_parameters.x_smart_token},
            json=profile,
        )
        if response.status_code != 201:
            adapter.error(
                "Failed to push profile to SmartRecruiters"
                " status_code={} response={}".format(
                    response.status_code,
                    response.text,
                )
            )
            failed_profiles.append(profile)
    return failed_profiles


def update_profiles(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
    parameters: WriteProfilesParameters,
    items: t.Iterable[t.Dict],
) -> t.List[t.Dict]:
    failed_profiles = []
    for profile in items:
        response = requests.patch(
            "{}/{}".format(SMARTRECRUITERS_PROFILES_ENDPOINT, profile["id"]),
            headers={"x-smarttoken": auth_parameters.x_smart_token},
            json=profile,
        )
        if response.status_code // 100 != 2:
            adapter.error(
                "Failed to update profile to SmartRecruiters profile_id={}"
                " status_code={} response={}".format(
                    profile["id"],
                    response.status_code,
                    response.text,
                )
            )
            failed_profiles.append(profile)
    return failed_profiles


def delete_profiles(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
    parameters: WriteProfilesParameters,
    items: t.Iterable[t.Dict],
) -> t.List[t.Dict]:
    failed_profiles = []
    for profile in items:
        response = requests.delete(
            "{}/{}".format(SMARTRECRUITERS_PROFILES_ENDPOINT, profile["id"]),
            headers={"x-smarttoken": auth_parameters.x_smart_token},
        )
        if response.status_code // 100 != 2:
            adapter.error(
                "Failed to delete profile from SmartRecruiters profile_id={}"
                " status_code={} response={}".format(
                    profile["id"],
                    response.status_code,
                    response.text,
                )
            )
            failed_profiles.append(profile)
    return failed_profiles


JobsAisle = Aisle(
    name=Entity.job,
    schema=SmartRecruitersJob,
    read=ReadOperation(
        criterias=Criterias(
            create=ReadJobsParameters,
            update=ReadJobsParameters,
            archive=ReadJobsParameters,
        ),
        function=merge(
            create=read_jobs,
            update=read_jobs,
            archive=read_jobs,
        ),
        # endpoints=Endpoints(
        #     create=GET_JOB_ENDPOINT,
        # ),
    ),
)

ProfilesAisle = Aisle(
    name=Entity.profile,
    schema=SmartRecruitersProfile,
    read=ReadOperation(
        criterias=Criterias(
            create=ReadProfilesParameters,
            update=ReadProfilesParameters,
            archive=ReadProfilesParameters,
        ),
        function=merge(
            create=read_profiles,
            update=read_profiles,
            archive=read_profiles,
        ),
        endpoints=Endpoints(
            create=GET_CANDIDATE_ENDPOINT,
            update=GET_CANDIDATE_ENDPOINT,
            archive=GET_CANDIDATE_ENDPOINT,
        ),
    ),
    write=WriteOperation(
        criterias=Criterias(
            create=WriteProfilesParameters,
            update=WriteProfilesParameters,
            archive=WriteProfilesParameters,
        ),
        function=merge(
            create=write_profiles,
            update=update_profiles,
            archive=delete_profiles,
        ),
        endpoints=Endpoints(
            create=POST_CANDIDATE_ENDPOINT,
            update=UPDATE_CANDIDATE_ENDPOINT,
            archive=DELETE_CANDIDATE_ENDPOINT,
        ),
    ),
)



import os
import typing as t
from pathlib import Path

from msgspec import DecodeError, Struct, convert, json
from pydantic import BaseModel

from hrflow_connectors.core.backend.common import BackendStore, msgspec_dec_hook

DIRECTORY_ENVIRONMENT_VARIABLE: t.Final = "HRFLOW_CONNECTORS_LOCALJSON_DIR"
STORE_FILENAME: t.Final = "store.json"
NAME: t.Final = "localjson"


class InternalState(Struct):
    path: Path


def empty_store() -> t.Dict:
    return dict(root="HrFlow Connectors", store=NAME, data=dict())


def get_state():
    directory = os.environ.get(DIRECTORY_ENVIRONMENT_VARIABLE)
    if directory is None:
        raise Exception(
            "Missing environment variable {} in order to setup LocalJson store".format(
                DIRECTORY_ENVIRONMENT_VARIABLE
            )
        )
    directory = Path(directory)
    if directory.is_absolute() is False:
        raise Exception(
            "{}={} should be an absolute filepath".format(
                DIRECTORY_ENVIRONMENT_VARIABLE, directory
            )
        )
    if directory.exists() is False:
        raise Exception(
            "{}={} does not exist".format(DIRECTORY_ENVIRONMENT_VARIABLE, directory)
        )
    path = directory / STORE_FILENAME
    if path.exists() is False:
        path.write_bytes(json.encode(empty_store()))
    else:
        try:
            json.decode(path.read_text())
        except (DecodeError, UnicodeDecodeError):
            raise Exception("Store file is corrupted. Unable to JSON decode")
    return InternalState(path=path)


def save(state: InternalState, key: str, data: t.Union[BaseModel, Struct]) -> None:
    store = json.decode(state.path.read_bytes())
    store["data"][key] = (
        data.json() if isinstance(data, BaseModel) else json.encode(data).decode()
    )
    state.path.write_bytes(json.encode(store))


def load(
    state: InternalState, key: str, parse_as: t.Union[type[BaseModel], type[Struct]]
):
    store = json.decode(state.path.read_bytes())
    if key in store["data"]:
        if issubclass(parse_as, BaseModel):
            return parse_as.parse_raw(store["data"][key])
        return convert(
            json.decode(store["data"][key]), parse_as, dec_hook=msgspec_dec_hook
        )
    return None


LocalJsonStore = BackendStore[InternalState](
    name=NAME, get_state=get_state, saver=save, loader=load
)



import typing as t
from enum import Enum

from msgspec import Struct


class Property(Struct):
    id: int
    internal: str
    apiKey: str
    value: str
    values: str


class candidateLocation(Struct):
    country: str
    lat: int
    lng: int
    city: str
    postalCode: str
    street: str
    streetNumber: str


class Candidate(Struct):
    id: int
    dateCreation: int
    firstName: str
    lastName: str
    mail: str
    phone: str
    cv: str
    lang: str
    socialLinks: t.List[t.Dict]
    unitId: int
    properties: t.Optional[t.List[t.Dict]]
    location: candidateLocation


class JobStatus(str, Enum):
    draft = "DRAFT"
    published = "PUBLISHED"
    done = "DONE"
    suspended = "SUSPENDED"


class ContractType(str, Enum):
    cdi = "CDI"
    cdd = "CDD"
    interim = "INTERIM"
    freelance = "FREELANCE"
    internship = "INTERNSHIP"
    apprenticeship = "APPRENTICESHIP"
    student = "STUDENT"
    vie = "VIE"
    franchise = "FRANCHISE"
    statute = "STATUTE"
    vacataire = "VACATAIRE"
    liberal = "LIBERAL"
    cdi_chantier = "CDI_CHANTIER"
    intermittent = "INTERMITTENT"
    season = "SEASON"
    other = "OTHER"
    volunteer = "VOLUNTEER"
    permenant = "PERMENANT"
    fixedterm = "FIXEDTERM"


class JobVisibility(str, Enum):
    public = "PUBLIC"
    internal = "INTERNAL"
    internal_and_public = "INTERNAL_AND_PUBLIC"
    private = "PRIVATE"


class Job(Struct):
    id: int
    token: str
    dateCreation: int
    dateFirstPublish: int
    dateLastPublish: int
    label: t.Optional[str]
    currentStatus: JobStatus
    contract: t.Optional[ContractType]
    contractLength: int
    fullTime: bool
    workHours: int
    remote: bool
    country: str
    city: str
    postalCode: str
    lat: str
    lng: str
    recruiterId: int
    who: str
    logo: str
    banner: str
    companyLabel: str
    tags: t.List[t.Dict]
    url: str
    urlApplying: str
    visibility: JobVisibility
    jobDescription: str
    profileDescription: str
    companyDescription: str
    properties: t.List[t.Dict]


class JobProperty(Struct):
    key: str
    value: str
    values: t.List[str]



from hrflow_connectors.v2.connectors.admen import Admen
from hrflow_connectors.v2.connectors.adzuna import Adzuna
from hrflow_connectors.v2.connectors.breezyhr import BreezyHR
from hrflow_connectors.v2.connectors.bullhorn import Bullhorn
from hrflow_connectors.v2.connectors.flatchr import Flatchr
from hrflow_connectors.v2.connectors.francetravail import FranceTravail
from hrflow_connectors.v2.connectors.hubspot import Hubspot
from hrflow_connectors.v2.connectors.recruitee import Recruitee
from hrflow_connectors.v2.connectors.smartrecruiters import SmartRecruiters
from hrflow_connectors.v2.connectors.taleez import Taleez
from hrflow_connectors.v2.connectors.teamtailor import Teamtailor
from hrflow_connectors.v2.connectors.waalaxy import Waalaxy
from hrflow_connectors.v2.connectors.workable import Workable
from hrflow_connectors.v2.connectors.zohorecruit import ZohoRecruit
from hrflow_connectors.v2.core.connector import (  # noqa: F401
    hrflow_connectors_manifest as hrflow_connectors_manifest,
)
from hrflow_connectors.v2.core.documentation import (  # noqa: F401
    hrflow_connectors_docs as hrflow_connectors_docs,
)

__CONNECTORS__ = [
    Bullhorn,
    ZohoRecruit,
    Admen,
    SmartRecruiters,
    Hubspot,
    Recruitee,
    Adzuna,
    FranceTravail,
    Flatchr,
    BreezyHR,
    Teamtailor,
    Taleez,
    Workable,
    Waalaxy,
]



import json
import time
import typing as t
from datetime import datetime
from io import BytesIO
from logging import LoggerAdapter

import requests
from msgspec import Meta, Struct
from typing_extensions import Annotated

from hrflow_connectors.v2.connectors.bullhorn.schemas import (
    BullhornJob,
    BullhornProfile,
)
from hrflow_connectors.v2.connectors.bullhorn.utils.authentication import auth
from hrflow_connectors.v2.core.common import Entity, Mode
from hrflow_connectors.v2.core.warehouse import (
    Aisle,
    Criterias,
    IncrementalTokenHandler,
    ReadOperation,
    WriteOperation,
    merge,
)


class AuthParameters(Struct):
    client_id: Annotated[
        str,
        Meta(
            description="Client identifier for Bullhorn",
        ),
    ]
    client_secret: Annotated[
        str,
        Meta(
            description="Client secret identifier for Bullhorn",
        ),
    ]
    password: Annotated[
        str,
        Meta(
            description="Password for Bullhorn login",
        ),
    ]
    username: Annotated[
        str,
        Meta(
            description="Username for Bullhorn login",
        ),
    ]


class UpdateApplicationsCriterias(Struct):
    job_id: Annotated[
        str,
        Meta(
            description="id for the job in Bullhorn",
        ),
    ]
    candidate_id: Annotated[
        str,
        Meta(
            description="id for of the profile in Bullhorn",
        ),
    ]
    # maybe should be optional
    status_when_created: Annotated[
        str,
        Meta(
            description="The status of the application when created in Bullhorn",
        ),
    ]
    # maybe should not be a parameters
    source: Annotated[
        str,
        Meta(
            description="The source of the application to be created in Bullhorn",
        ),
    ]


class BaseParameters(Struct):
    limit: Annotated[
        t.Optional[int],
        Meta(
            description="Number of items to pull, ignored if not provided.",
        ),
    ] = None


class BaseJobsParameters(BaseParameters, kw_only=True):
    fields: Annotated[
        str,
        Meta(
            min_length=2,
            description="List of job fields to be retrieved from Bullhorn",
        ),
    ] = (
        "address,assignedUsers,businessSectors,categories,clientBillRate,"
        "clientContact,clientCorporation,costCenter,customInt1,customInt2,"
        "customText1,customText10,customText11,customText12,customText13,"
        "customText2,customText3,customText4,customText5,customText6,"
        "customText7,customText8,customText9,customTextBlock1,customTextBlock2,"
        "customTextBlock3,customTextBlock4,customTextBlock5,dateAdded,dateEnd,"
        "degreeList,description,publicDescription,durationWeeks,educationDegree,"
        "employmentType,feeArrangement,hoursOfOperation,hoursPerWeek,isOpen,"
        "isWorkFromHome,markUpPercentage,numOpenings,onSite,payRate,salary,salaryUnit,"
        "skills,skillList,source,specialties,startDate,status,title,id,type,willRelocate,"
        "owner"
    )
    query: Annotated[
        str,
        Meta(
            description=(
                "This query will restrict the results retrieved from Bullhorn based on"
                " the specified conditions"
            ),
        ),
    ] = "isDeleted:0 AND isOpen:true"


class ReadCreatedJobsCriterias(BaseJobsParameters, kw_only=True):
    created_date: Annotated[
        datetime,
        Meta(
            description="The creation date from which you want to pull jobs",
        ),
    ]


class ReadUpdatedJobsCriterias(BaseJobsParameters, kw_only=True):
    last_modified_date: Annotated[
        datetime,
        Meta(
            description="The modification date from which you want to pull jobs",
        ),
    ]


class ReadArchivedJobsCriterias(BaseParameters, kw_only=True):
    last_modified_date: Annotated[
        datetime,
        Meta(
            description=(
                "The modification date from which you want to pull jobs and archive"
                " them"
            ),
        ),
    ]
    query: Annotated[
        str,
        Meta(
            description=(
                "This query will restrict the results retrieved from Bullhorn based on"
                " the specified conditions"
            ),
        ),
    ] = "isDeleted:0 AND isOpen:true"
    fields: Annotated[
        str,
        Meta(
            description="Field to be used as reference for archiving",
        ),
    ] = "id"


class BaseProfilesParameters(BaseParameters):
    fields: Annotated[
        str,
        Meta(
            min_length=2,
            description="List of profile fields to be retrieved from Bullhorn",
        ),
    ] = (
        "address,businessSectors,categories,companyName,customInt4,customInt5,"
        "customInt6,customText1,customText10,customText11,customText12,"
        "customText13,customText14,customText15,customText16,customText18,"
        "customText23,customText24,customText25,customText4,customText5,"
        "customText6,customText9,dateAdded,dateAvailable,dateAvailableEnd,"
        "dateLastModified,dateOfBirth,dayRate,dayRateLow,degreeList,"
        "desiredLocations,description,disability,educations,email,email2,"
        "employmentPreference,ethnicity,experience,firstName,id,lastName,"
        "mobile,name,namePrefix,occupation,owner,phone,primarySkills,"
        "secondaryOwners,secondarySkills,salary,salaryLow,skillSet,"
        "source,specialties,status,userDateAdded,veteran,willRelocate,"
        "workHistories,workPhone"
    )
    query: Annotated[
        str,
        Meta(
            description=(
                "This query will restrict the results retrieved from Bullhorn based on"
                " the specified conditions"
            ),
        ),
    ] = "isDeleted:0"


class ReadCreatedProfilesCriterias(BaseProfilesParameters, kw_only=True):
    created_date: Annotated[
        datetime,
        Meta(
            description="The creation date from which you want to pull profiles",
        ),
    ]
    parse_resume: Annotated[
        bool,
        Meta(
            description=(
                "If True, resumes will be retrieved and parsed along with the profile"
                " data"
            ),
        ),
    ] = False


class ReadUpdatedProfilesCriterias(BaseProfilesParameters, kw_only=True):
    last_modified_date: Annotated[
        datetime,
        Meta(
            description="The modification date from which you want to pull profiles",
        ),
    ]
    parse_resume: Annotated[
        bool,
        Meta(
            description=(
                "If True, resumes will be retrieved and parsed along with the profile"
                " data"
            ),
        ),
    ] = False


class ReadArchivedProfilesCriterias(BaseParameters, kw_only=True):
    last_modified_date: Annotated[
        datetime,
        Meta(
            description="The modification date from which you want to pull profiles",
        ),
    ]
    query: Annotated[
        str,
        Meta(
            description=(
                "This query will restrict the results retrieved from Bullhorn based on"
                " the specified conditions"
            ),
        ),
    ] = "isDeleted:1"
    fields: Annotated[
        str,
        Meta(
            description="Field to be used as reference for archiving",
        ),
    ] = "id"


def make_request(
    method, url, params, auth_parameters: AuthParameters, adapter, json=None
) -> t.Optional[dict]:
    response = method(url, params=params, data=json)
    if response.status_code == 401:
        adapter.info("Auth token expired, regenerating...")
        auth_info = auth(
            auth_parameters.username,
            auth_parameters.password,
            auth_parameters.client_id,
            auth_parameters.client_secret,
        )
        params["BhRestToken"] = auth_info["BhRestToken"]
        response = method(url, params=params, data=json)
    return handle_response(response, adapter)


def handle_response(response, adapter) -> t.Optional[dict]:
    if not response.ok:
        adapter.error(
            f"Request failed with status_code={response.status_code},"
            f" response={response.text}"
        )
        return None
    return response.json()


def search_entity(
    entity, rest_url, bh_rest_token, query, fields, adapter, auth_parameters
):
    search_url = f"{rest_url}search/{entity}"
    params = {
        "BhRestToken": bh_rest_token,
        "query": query,
        "fields": fields,
        "sort": "id",
    }
    response = make_request(requests.get, search_url, params, auth_parameters, adapter)
    return response


def create_entity(
    entity: str,
    rest_url: str,
    params: dict,
    data: dict,
    auth_parameters: AuthParameters,
    adapter: LoggerAdapter,
) -> t.Optional[dict]:
    url = f"{rest_url}entity/{entity}"
    response = make_request(
        requests.post, url, params, auth_parameters, adapter, json.dumps(data)
    )
    return response


def update_entity(
    entity: str,
    entity_id: str,
    rest_url: str,
    params: dict,
    data: dict,
    auth_parameters: AuthParameters,
    adapter: LoggerAdapter,
) -> t.Optional[dict]:
    url = f"{rest_url}entity/{entity}/{entity_id}"
    response = make_request(
        requests.put, url, params, auth_parameters, adapter, json.dumps(data)
    )
    return response


def check_entity_files(
    entity: str,
    entity_id: str,
    rest_url: str,
    params: dict,
    auth_parameters: AuthParameters,
    adapter: LoggerAdapter,
) -> t.Optional[dict]:
    url = f"{rest_url}entityFiles/{entity}/{entity_id}"
    response = make_request(requests.get, url, params, auth_parameters, adapter)
    return response


def upload_attachment(
    entity: str,
    entity_id: str,
    rest_url: str,
    params: dict,
    auth_parameters: AuthParameters,
    adapter: LoggerAdapter,
    attachment,
) -> t.Optional[dict]:
    url = f"{rest_url}file/{entity}/{entity_id}"
    attachment_response = make_request(
        requests.put, url, params, auth_parameters, adapter, json.dumps(attachment)
    )
    return attachment_response


def update_application(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
    parameters: UpdateApplicationsCriterias,
    items: t.Iterable[dict],
) -> list[dict]:
    failed_profiles = []
    auth_info = auth(
        auth_parameters.username,
        auth_parameters.password,
        auth_parameters.client_id,
        auth_parameters.client_secret,
    )
    rest_url = auth_info["restUrl"]
    bh_rest_token = auth_info["BhRestToken"]
    params = {"BhRestToken": bh_rest_token}

    for profile in items:
        if parameters.candidate_id:
            candidate_id = parameters.candidate_id

        if not parameters.candidate_id:
            attachment = profile.pop("attachment", None)
            profile["source"] = [parameters.source or profile.get("source")]
            profile["status"] = parameters.status_when_created or profile.get("status")
            email = profile["email"]

            adapter.info(f"Checking if candidate with email: {email} already exists.")
            search_results = search_entity(
                "Candidate",
                rest_url,
                bh_rest_token,
                f"(email:{email} OR email2:{email}) AND isDeleted:0",
                (
                    "id,isDeleted,dateAdded,status,source,email,firstName,"
                    "lastName,name,mobile,address"
                ),
                adapter,
                parameters,
            )

            if not search_results:
                failed_profiles.append(profile)
                continue

            if search_results["count"] == 0:
                adapter.info(f"Creating candidate with email: {email}")
                candidate_response = create_entity(
                    "Candidate", rest_url, params, profile, auth_parameters, adapter
                )
                if not candidate_response:
                    failed_profiles.append(profile)
                    continue
                candidate_id = candidate_response["changedEntityId"]
                attachment_exists = False

            else:
                candidate_data = search_results["data"][0]
                candidate_id = candidate_data.get("id")

                profile.update(
                    {
                        "firstName": candidate_data.get("firstName") or profile.get(
                            "firstName"
                        ),
                        "lastName": candidate_data.get("lastName") or profile.get(
                            "lastName"
                        ),
                        "name": candidate_data.get("name") or profile.get("name"),
                        "address": candidate_data.get("address") or profile.get(
                            "address"
                        ),
                        "mobile": candidate_data.get("mobile") or profile.get("mobile"),
                        "status": candidate_data.get("status") or profile.get("status"),
                        "source": candidate_data.get("source") or profile.get("source"),
                    }
                )

                adapter.info(f"Updating candidate with ID: {candidate_id}")
                candidate_response = update_entity(
                    "Candidate",
                    candidate_id,
                    rest_url,
                    params,
                    profile,
                    auth_parameters,
                    adapter,
                )

                if not candidate_response:
                    failed_profiles.append(profile)
                    continue

                if attachment:
                    adapter.info(
                        f"Checking if attachment exists for candidate {candidate_id}"
                    )
                    entity_files = check_entity_files(
                        "Candidate",
                        candidate_id,
                        rest_url,
                        params,
                        auth_parameters,
                        adapter,
                    )
                    if entity_files:
                        attachment_exists = any(
                            file["name"] == attachment["name"]
                            for file in entity_files.get("EntityFiles", [])
                        )

            if not attachment_exists:
                adapter.info("Uploading attachment")
                attachment_response = upload_attachment(
                    "Candidate",
                    candidate_id,
                    rest_url,
                    params,
                    auth_parameters,
                    adapter,
                    attachment,
                )
                if not attachment_response:
                    failed_profiles.append(profile)
                    continue

        adapter.info(
            f"Checking if candidate {candidate_id} has already applied for job"
            f" {parameters.job_id}"
        )
        job_submission_results = search_entity(
            "JobSubmission",
            rest_url,
            bh_rest_token,
            f"candidate.id:{candidate_id} AND jobOrder.id:{parameters.job_id}",
            "id,status,dateAdded",
            adapter,
            parameters,
        )

        if job_submission_results:
            job_submission_exists = job_submission_results.get("count", 0) > 0

            job_submission_payload = {
                "candidate": {"id": candidate_id},
                "jobOrder": {"id": parameters.job_id},
                "status": parameters.status_when_created,
                "dateWebResponse": int(time.time() * 1000),
            }
            if job_submission_exists:
                job_submission_id = job_submission_results["data"][0]["id"]
                adapter.info("Updating JobSubmission")
                job_submission_response = update_entity(
                    "JobSubmission",
                    job_submission_id,
                    rest_url,
                    params,
                    job_submission_payload,
                    auth_parameters,
                    adapter,
                )
            else:
                adapter.info("Creating JobSubmission")
                job_submission_response = create_entity(
                    "JobSubmission",
                    rest_url,
                    params,
                    job_submission_payload,
                    auth_parameters,
                    adapter,
                )

        if not job_submission_response:
            failed_profiles.append(profile)

    return failed_profiles


def generic_job_pulling(
    action: Mode,
):
    def _pull_items(
        adapter: LoggerAdapter,
        auth_parameters: AuthParameters,
        parameters: t.Union[
            ReadCreatedJobsCriterias,
            ReadUpdatedJobsCriterias,
            ReadArchivedJobsCriterias,
        ],
        incremental: bool,
        incremental_token: t.Optional[str],
    ) -> t.Iterable[dict]:
        start, auth_retries, total_returned = 0, 0, 0
        should_break = False

        authentication = auth(
            auth_parameters.username,
            auth_parameters.password,
            auth_parameters.client_id,
            auth_parameters.client_secret,
        )

        if action is Mode.create:
            date_field = "created_date"
            bullhorn_date_field = "dateAdded"
        else:
            date_field = "last_modified_date"
            bullhorn_date_field = "dateLastModified"

        last_id = ""
        if not incremental:
            date_value = getattr(parameters, date_field)
            if date_value is None:
                raise Exception(f"{date_field} cannot be None in ReadMode.sync")
            last_id = None
        else:
            if incremental_token:
                try:
                    read_data = json.loads(incremental_token)
                    date_value = read_data[date_field]
                    last_id = read_data["last_id"]
                except (json.JSONDecodeError, KeyError) as e:
                    raise Exception(f"Error parsing read_from: {e}")
            else:
                date_value = getattr(parameters, date_field)
                last_id = None

        date_filter = transform_iso(date_value)
        if not date_filter:
            raise Exception(f"Error applying transformation on {date_field}")

        # Construct the query
        query = f"{bullhorn_date_field}:[{date_filter} TO *]"
        if parameters.query:
            query = f"{parameters.query} AND {query}"
        # Fetch and process jobs
        while True:
            try:
                jobs_url = f"{authentication['restUrl']}search/JobOrder"
                params = {
                    "query": query,
                    "fields": parameters.fields,
                    "sort": f"{bullhorn_date_field},id",
                    "start": str(start),
                }
                if parameters.limit:
                    params["count"] = str(parameters.limit)

                headers = {"BhRestToken": authentication["BhRestToken"]}
                response = requests.get(url=jobs_url, params=params, headers=headers)
                if response.status_code // 100 != 2:
                    adapter.error(
                        "Failed to pull jobs from Bullhorn"
                        f" status_code={response.status_code} response={response.text}"
                    )
                    raise Exception("Failed to pull jobs from Bullhorn")

                response = response.json()
                start = response["start"] + response["count"]
                data = response["data"]

                for job in data:
                    if parameters.limit and total_returned >= parameters.limit:
                        should_break = True
                        break

                    date_added = job.get("dateAdded")
                    date_last_modified = job.get("dateLastModified")

                    transformed_date_added = transform_timestamp_read_from(date_added)
                    transformed_date_last_modified = transform_timestamp_read_from(
                        date_last_modified
                    )

                    if (
                        action is Mode.create
                        and transformed_date_added is not None
                        and transformed_date_last_modified is not None
                        and transformed_date_added[:19]
                        != transformed_date_last_modified[
                            :19
                        ]  # ignore microsecond difference created by Bullhorn
                    ) or (action is Mode.update and date_added == date_last_modified):
                        continue

                    if (
                        last_id
                        and job[bullhorn_date_field] == date_value
                        and job["id"] <= last_id
                    ):
                        adapter.info("Skipping job with id <= last_id")
                        continue
                    yield job
                    total_returned += 1

                if should_break:
                    break

                if start >= response["total"]:
                    break

            except requests.HTTPError as e:
                if e.response.status_code == 401:
                    adapter.info("Received 401 error. Retrying authentication.")
                    if auth_retries > 2:
                        raise Exception("Max auth retries exceeded")
                    authentication = auth(
                        auth_parameters.username,
                        auth_parameters.password,
                        auth_parameters.client_id,
                        auth_parameters.client_secret,
                        refresh_token=authentication["refresh_token"],
                    )
                    auth_retries += 1
                else:
                    adapter.error("Failed to fetch jobs from Bullhorn.")
                    raise e

    return _pull_items


def generic_profile_pulling(
    action: Mode,
):
    def __pull_items(
        adapter: LoggerAdapter,
        auth_parameters: AuthParameters,
        parameters: t.Union[
            ReadCreatedProfilesCriterias,
            ReadUpdatedProfilesCriterias,
            ReadArchivedProfilesCriterias,
        ],
        incremental: bool,
        incremental_token: t.Optional[str],
    ) -> t.Iterable[dict]:
        authentication = auth(
            auth_parameters.username,
            auth_parameters.password,
            auth_parameters.client_id,
            auth_parameters.client_secret,
        )
        start, auth_retries, total_returned = 0, 0, 0
        should_break = False

        authentication = auth(
            auth_parameters.username,
            auth_parameters.password,
            auth_parameters.client_id,
            auth_parameters.client_secret,
        )
        if action is Mode.create:
            date_field = "created_date"
            bullhorn_date_field = "dateAdded"
        else:
            date_field = "last_modified_date"
            bullhorn_date_field = "dateLastModified"

        last_id = ""
        if not incremental:
            date_value = getattr(parameters, date_field)
            if date_value is None:
                raise Exception(f"{date_field} cannot be None in ReadMode.sync")
            last_id = None
        else:
            if incremental_token:
                try:
                    read_data = json.loads(incremental_token)
                    date_value = read_data[date_field]
                    last_id = read_data["last_id"]
                except (json.JSONDecodeError, KeyError) as e:
                    raise Exception(f"Error parsing read_from: {e}")
            else:
                date_value = getattr(parameters, date_field)
                last_id = None

        date_filter = transform_iso(date_value)
        if not date_filter:
            raise Exception(f"Error applying transformation on {date_field}")

        # Construct the query
        query = f"{bullhorn_date_field}:[{date_filter} TO *]"
        if parameters.query:
            query = f"{parameters.query} AND {query}"

        while True:
            try:
                profiles_url = f"{authentication['restUrl']}search/Candidate"
                params = {
                    "query": query,
                    "fields": parameters.fields,
                    "sort": f"{bullhorn_date_field},id",
                    "start": str(start),
                }

                if parameters.limit:
                    params["count"] = str(parameters.limit)

                headers = {"BhRestToken": authentication["BhRestToken"]}

                response = requests.get(
                    url=profiles_url, params=params, headers=headers
                )
                if response.status_code // 100 != 2:
                    adapter.error(
                        "Failed to pull profiles from Bullhorn"
                        f" status_code={response.status_code} response={response.text}"
                    )
                    raise Exception("Failed to pull profiles from Bullhorn")
                response = response.json()

                start = response["start"] + response["count"]
                total = response["total"]
                data = response["data"]

                for profile in data:
                    if parameters.limit and total_returned >= parameters.limit:
                        should_break = True
                        break

                    date_added = profile.get("dateAdded")
                    date_last_modified = profile.get("dateLastModified")

                    transformed_date_added = transform_timestamp_read_from(date_added)
                    transformed_date_last_modified = transform_timestamp_read_from(
                        date_last_modified
                    )

                    if (
                        action is Mode.create
                        and transformed_date_added is not None
                        and transformed_date_last_modified is not None
                        and transformed_date_added[:19]
                        != transformed_date_last_modified[
                            :19
                        ]  # ignore microsecond difference created by Bullhorn
                    ) or (action is Mode.update and date_added == date_last_modified):
                        continue

                    if (
                        last_id
                        and profile.get(bullhorn_date_field) == date_value
                        and profile.get("id") <= last_id
                    ):
                        adapter.info("Skipping profile with id <= last_id")
                        continue

                    if action is Mode.archive:
                        yield profile
                        total_returned += 1
                        continue

                    if (
                        not isinstance(parameters, ReadArchivedProfilesCriterias)
                        and parameters.parse_resume
                    ):
                        profile["cvFile"] = None
                        url_files = (
                            authentication["restUrl"]
                            + "entityFiles/Candidate/"
                            + str(profile["id"])
                        )
                        headers = {"BhRestToken": authentication["BhRestToken"]}
                        response = requests.get(url=url_files, headers=headers)
                        response = response.json()

                        last_cv = None
                        curr_entity = None
                        file_name = None
                        if len(response["EntityFiles"]) > 0:
                            for entity_file in response["EntityFiles"]:
                                if entity_file["type"] == "Resume":
                                    if not curr_entity:
                                        curr_entity = entity_file
                                        last_cv = entity_file["id"]
                                        file_name = entity_file["name"]
                                    elif (
                                        curr_entity["dateAdded"]
                                        < entity_file["dateAdded"]
                                    ):
                                        curr_entity = entity_file
                                        last_cv = entity_file["id"]
                                        file_name = entity_file["name"]

                        if last_cv is not None:
                            url_cv = (
                                authentication["restUrl"]
                                + "/file/Candidate/"
                                + str(profile["id"])
                                + "/"
                                + str(last_cv)
                                + "/raw"
                            )
                            response = requests.get(url=url_cv, headers=headers)
                            file = response.content
                            profile_file = BytesIO(file)
                            profile["cvFile"] = profile_file
                            profile["fileName"] = file_name

                    if "educations" in parameters.fields:
                        education_ids = []
                        educations = []
                        for ed in profile["educations"]["data"]:
                            education_ids.append(ed["id"])
                        for id in education_ids:
                            education_url = (
                                f"{authentication['restUrl']}entity/CandidateEducation"
                                f"/{str(id)}?fields='city,school,startDate,endDate,"
                                "degree,certification,comments'"
                            )

                            response = requests.get(url=education_url, headers=headers)
                            response = response.json()
                            educations.append(response["data"])

                        profile["educations"] = educations

                    if "workHistories" in parameters.fields:
                        work_history_ids = []
                        work_histories = []
                        for work_history in profile["workHistories"]["data"]:
                            work_history_ids.append(work_history["id"])
                        for id in work_history_ids:
                            work_history_url = (
                                f"{authentication['restUrl']}entity/"
                                f"CandidateWorkHistory/{str(id)}"
                                "?fields='title,comments,startDate,endDate,companyName'"
                            )
                            response = requests.get(
                                url=work_history_url, headers=headers
                            )
                            response = response.json()
                            work_histories.append(response["data"])

                        profile["workHistories"] = work_histories

                    total_returned += 1
                    yield profile

                if should_break:
                    break

                if start >= total:
                    break

            except requests.HTTPError as e:
                if e.response.status_code == 401:
                    adapter.info(
                        "Received 401 error. Retrying authentication to continue"
                        " fetching profiles."
                    )
                    if auth_retries > 2:
                        raise Exception(
                            f"Retries exceeded for authentication ({auth_retries})."
                            " Stopping execution."
                        )

                    authentication = auth(
                        auth_parameters.username,
                        auth_parameters.password,
                        auth_parameters.client_id,
                        auth_parameters.client_secret,
                        refresh_token=authentication["refresh_token"],
                    )
                    auth_retries += 1
                    continue
                else:
                    adapter.error("Failed to fetch profiles from Bullhorn.")
                    raise e

    return __pull_items


def transform_iso(iso_date: t.Optional[t.Union[str, datetime]]) -> t.Optional[str]:
    if iso_date is None:
        return None

    if isinstance(iso_date, str):
        dt = datetime.fromisoformat(iso_date.replace("Z", "+00:00"))
    elif isinstance(iso_date, datetime):
        dt = iso_date
    else:
        raise TypeError(f"Expected str or datetime, got {type(iso_date)}")

    # Return the date formatted in the desired format
    return dt.strftime("%Y%m%d%H%M%S")


def transform_timestamp_read_from(
    timestamp: t.Optional[t.Union[float, int]],
) -> t.Optional[str]:
    if timestamp is None:
        return None
    transformed_date = datetime.utcfromtimestamp(int(timestamp) / 1000)
    return transformed_date.isoformat()


def item_to_read_from_create(item: dict) -> str:
    created_date = transform_timestamp_read_from(item["dateAdded"])
    return json.dumps(dict(created_date=created_date, last_id=item["id"]))


def item_to_read_from_update_or_archive(item: dict) -> str:
    last_modified_date = transform_timestamp_read_from(item["dateLastModified"])
    return json.dumps(dict(last_modified_date=last_modified_date, last_id=item["id"]))


ProfilesAisle = Aisle(
    name=Entity.profile,
    read=ReadOperation(
        function=merge(
            create=generic_profile_pulling(Mode.create),
            update=generic_profile_pulling(Mode.update),
            archive=generic_profile_pulling(Mode.archive),
        ),
        criterias=Criterias(
            create=ReadCreatedProfilesCriterias,
            update=ReadUpdatedProfilesCriterias,
            archive=ReadArchivedProfilesCriterias,
        ),
        incremental_token_handler=IncrementalTokenHandler(
            create=item_to_read_from_create,
            update=item_to_read_from_update_or_archive,
            archive=item_to_read_from_update_or_archive,
        ),
    ),
    schema=BullhornProfile,
)

JobsAisle = Aisle(
    name=Entity.job,
    read=ReadOperation(
        function=merge(
            create=generic_job_pulling(Mode.create),
            update=generic_job_pulling(Mode.update),
            archive=generic_job_pulling(Mode.archive),
        ),
        criterias=Criterias(
            create=ReadCreatedJobsCriterias,
            update=ReadUpdatedJobsCriterias,
            archive=ReadArchivedJobsCriterias,
        ),
        incremental_token_handler=IncrementalTokenHandler(
            create=item_to_read_from_create,
            update=item_to_read_from_update_or_archive,
            archive=item_to_read_from_update_or_archive,
        ),
    ),
    schema=BullhornJob,
)


ApplicationsAisle = Aisle(
    name=Entity.application,
    write=WriteOperation(
        function=merge(update=update_application),
        criterias=Criterias(update=UpdateApplicationsCriterias),
    ),
    schema=BullhornProfile,
)



import typing as t
from logging import LoggerAdapter

from hrflow import Hrflow
from pydantic import Field

from hrflow_connectors.core import (
    DataType,
    FieldType,
    ParametersModel,
    ReadMode,
    Warehouse,
    WarehouseReadAction,
    WarehouseWriteAction,
)
from hrflow_connectors.v1.connectors.hrflow.schemas import (
    HrFlowProfile,
    HrFlowProfileParsing,
)


class ReadProfileParameters(ParametersModel):
    api_secret: str = Field(
        ...,
        description="X-API-KEY used to access HrFlow.ai API",
        repr=False,
        field_type=FieldType.Auth,
    )
    api_user: str = Field(
        ...,
        description="X-USER-EMAIL used to access HrFlow.ai API",
        field_type=FieldType.Auth,
    )
    source_key: str = Field(
        ..., description="HrFlow.ai source key", field_type=FieldType.QueryParam
    )
    profile_key: str = Field(
        ..., description="HrFlow.ai profile key", field_type=FieldType.QueryParam
    )


class WriteProfileParameters(ParametersModel):
    api_secret: str = Field(
        ...,
        description="X-API-KEY used to access HrFlow.ai API",
        repr=False,
        field_type=FieldType.Auth,
    )
    api_user: str = Field(
        ...,
        description="X-USER-EMAIL used to access HrFlow.ai API",
        field_type=FieldType.Auth,
    )
    source_key: str = Field(
        ..., description="HrFlow.ai source key", field_type=FieldType.QueryParam
    )
    edit: bool = Field(
        False,
        description="When enabled the profile must exist in the source",
        field_type=FieldType.Other,
    )
    only_edit_fields: t.Optional[t.List[str]] = Field(
        ...,
        description=(
            "List of attributes to use for the edit operation e.g. ['tags',"
            " 'metadatas']"
        ),
        field_type=FieldType.Other,
    )


class WriteProfileParsingParameters(ParametersModel):
    api_secret: str = Field(
        ...,
        description="X-API-KEY used to access HrFlow.ai API",
        repr=False,
        field_type=FieldType.Auth,
    )
    api_user: str = Field(
        ...,
        description="X-USER-EMAIL used to access HrFlow.ai API",
        field_type=FieldType.Auth,
    )
    source_key: str = Field(
        ..., description="HrFlow.ai source key", field_type=FieldType.Other
    )
    only_insert: bool = Field(
        False,
        description=(
            "When enabled the profile is written only if it doesn't exist in the source"
        ),
        field_type=FieldType.Other,
    )


def read(
    adapter: LoggerAdapter,
    parameters: ReadProfileParameters,
    read_mode: t.Optional[ReadMode] = None,
    read_from: t.Optional[str] = None,
) -> t.List[t.Dict]:
    hrflow_client = Hrflow(
        api_secret=parameters.api_secret, api_user=parameters.api_user
    )
    response = hrflow_client.profile.storing.get(
        source_key=parameters.source_key, key=parameters.profile_key
    )
    if "Unable to find object" in response["message"]:
        adapter.info(
            "No profile found for source_key={} profile_key={} response={}".format(
                parameters.source_key, parameters.profile_key, response
            )
        )
        return []
    elif response["code"] >= 400:
        adapter.error(
            "Failed to get profile source_key={} profile_key={} response={}".format(
                parameters.source_key, parameters.profile_key, response
            )
        )
        raise Exception("Failed to get profile")
    return [response["data"]]


def write(
    adapter: LoggerAdapter,
    parameters: WriteProfileParameters,
    profiles: t.Iterable[t.Dict],
) -> t.List[t.Dict]:
    failed = []
    hrflow_client = Hrflow(
        api_secret=parameters.api_secret, api_user=parameters.api_user
    )
    for profile in profiles:
        if parameters.edit:
            current_profile = hrflow_client.profile.storing.get(
                source_key=parameters.source_key, reference=profile["reference"]
            ).get("data")
            if not current_profile:
                adapter.warning(
                    "Mode edit is activated and profile with reference={} not found"
                    " in source. Failing for profile...".format(profile["reference"])
                )
                failed.append(profile)
                continue
            if parameters.only_edit_fields:
                edit = {
                    field: profile.get(field) for field in parameters.only_edit_fields
                }
            else:
                edit = profile
            profile_to_index = {**current_profile, **edit}

            response = hrflow_client.profile.storing.edit(
                source_key=parameters.source_key,
                key=current_profile["key"],
                profile_json=profile_to_index,
            )
            if response["code"] != 200:
                adapter.error(
                    "Failed to edit profile with reference={} key={} response={}"
                    .format(
                        profile_to_index["key"], profile_to_index["reference"], response
                    )
                )
                failed.append(profile)
        else:
            response = hrflow_client.profile.storing.add_json(
                source_key=parameters.source_key, profile_json=profile
            )
            if response["code"] // 100 != 2:
                adapter.error("Failed to add profile with response={}".format(response))
                failed.append(profile)
    return failed


def merge_info(base: dict, info: dict) -> dict:
    if not info:
        return base

    info_parsed = base.get("info", {})
    existing_urls = info_parsed.get("urls", [])

    if isinstance(info.get("urls"), list):
        for new_url in info["urls"]:
            if new_url not in existing_urls:
                existing_urls.append(new_url)

    info_parsed["urls"] = existing_urls

    for key, value in info.items():
        if value and key != "location" and key != "urls":
            info_parsed[key] = value
        elif key == "location" and isinstance(value, dict) and any(value.values()):
            info_parsed[key] = value

    base["info"] = info_parsed
    return base


def merge_item(base: dict, profile: dict, item: str) -> dict:
    if not profile.get(item):
        return base

    base[item] = profile[item]
    return base


def hydrate_profile(profile_parsed: dict, profile_json: dict) -> dict:
    profile_info = profile_json.get("info", {})
    profile_enriched = merge_info(profile_parsed, profile_info)

    items_to_merge = [
        "experiences",
        "educations",
        "skills",
        "languages",
        "certifications",
        "interests",
    ]
    for item in items_to_merge:
        profile_enriched = merge_item(profile_enriched, profile_json, item)

    profile_enriched["text"] = profile_json.get("text") or profile_enriched.get("text")
    profile_enriched["text_language"] = profile_json.get(
        "text_language"
    ) or profile_enriched.get("text_language")
    profile_enriched["experiences_duration"] = (
        profile_json.get("experiences_duration")
        if profile_json.get("experiences_duration") is not None
        else profile_enriched.get("experiences_duration")
    )
    profile_enriched["educations_duration"] = (
        profile_json.get("educations_duration")
        if profile_json.get("educations_duration") is not None
        else profile_enriched.get("educations_duration")
    )
    return profile_enriched


def write_parsing(
    adapter: LoggerAdapter,
    parameters: WriteProfileParsingParameters,
    profiles: t.Iterable[t.Dict],
) -> t.List[t.Dict]:
    failed = []
    hrflow_client = Hrflow(
        api_secret=parameters.api_secret, api_user=parameters.api_user
    )

    source_response = hrflow_client.source.get(key=parameters.source_key)

    for profile in profiles:
        if parameters.only_insert and hrflow_client.profile.storing.get(
            source_key=parameters.source_key, reference=profile["reference"]
        ).get("data"):
            adapter.info(
                "Mode only_insert is activated. Profile with reference={} already"
                " in source. Skipping...".format(profile["reference"])
            )
            continue

        if profile.get("resume") is None:
            indexing_response = hrflow_client.profile.storing.add_json(
                source_key=parameters.source_key, profile_json=profile
            )
            if indexing_response["code"] != 201:
                adapter.error(
                    "Failed to index profile with reference={} response={}".format(
                        profile["reference"], indexing_response
                    )
                )
                failed.append(profile)
            continue

        parsing_response = hrflow_client.profile.parsing.add_file(
            source_key=parameters.source_key,
            profile_file=profile["resume"]["raw"],
            profile_content_type=profile["resume"]["content_type"],
            reference=profile["reference"],
            tags=profile["tags"],
            metadatas=profile["metadatas"],
            created_at=profile["created_at"],
        )
        if parsing_response["code"] not in [202, 201]:  # 202: Accepted, 201: Created
            adapter.error(
                "Failed to parse profile with reference={} response={}".format(
                    profile["reference"], parsing_response
                )
            )
            failed.append(profile)
            continue

        # check if sync_parsing is enabled
        if source_response["code"] != 200:
            adapter.warning(
                "Failed to get source with key={} response={}, won't be able to update"
                " profile parsed with profile json".format(
                    parameters.source_key, source_response
                )
            )
        elif source_response["data"]["sync_parsing"] is True:
            current_profile = parsing_response["data"]["profile"]
            profile_result = hydrate_profile(current_profile, profile)

            edit_response = hrflow_client.profile.storing.edit(
                source_key=parameters.source_key,
                key=profile_result["key"],
                profile_json=profile_result,
            )
            if edit_response["code"] != 200:
                adapter.warning(
                    "Failed to update profile after parsing, reference={}"
                    " response={}".format(profile["reference"], edit_response)
                )
                failed.append(profile)
                continue

    return failed


HrFlowProfileWarehouse = Warehouse(
    name="HrFlow.ai Profiles",
    data_schema=HrFlowProfile,
    data_type=DataType.profile,
    read=WarehouseReadAction(parameters=ReadProfileParameters, function=read),
    write=WarehouseReadAction(parameters=WriteProfileParameters, function=write),
)


HrFlowProfileParsingWarehouse = Warehouse(
    name="HrFlow.ai Profile Parsing",
    data_schema=HrFlowProfileParsing,
    data_type=DataType.profile,
    write=WarehouseWriteAction(
        parameters=WriteProfileParsingParameters, function=write_parsing
    ),
)



import typing as t
from enum import Enum

from msgspec import Struct
from pydantic import BaseModel

Schema = t.Union[type[BaseModel], type[Struct]]
Parameters = t.Union[Struct, BaseModel]


class Entity(Enum):
    job = "job"
    profile = "profile"
    application = "application"


class Mode(Enum):
    create = "create"
    update = "update"
    archive = "archive"


class Direction(Enum):
    inbound = "inbound"
    outbound = "outbound"



import typing as t

from hrflow_connectors.v2.connectors.zohorecruit.warehouse import ZohoRecruitWarehouse
from hrflow_connectors.v2.core.common import Direction, Entity, Mode
from hrflow_connectors.v2.core.connector import Connector, ConnectorType, Flow


def get_location(record: dict) -> dict:
    parts = []

    street = record.get("Street", None)
    city = record.get("City", None)
    state = record.get("State", None)
    country = record.get("Country", None)
    zip_code = record.get("Zip_Code", None)
    parts = [street, city, state, country, zip_code]

    location_text = " ".join([part for part in parts if part])

    location = dict(
        text=location_text,
        lat=None,
        lng=None,
        fields=dict(
            postcode=zip_code,
            city=city,
            state=state,
            country=country,
            text=location_text,
        ),
    )
    return location


def get_job_opening_tags(job_opening: dict) -> list:
    tags_list = [
        "AccountManager",
        "AssignedRecruiter",
        "ClientName",
        "ContactName",
        "Date_Opened",
        "Industry",
        "Job_Opening_Status",
        "JobType",
        "Number_of_Positions",
        "No_of_Candidates_Associated",
        "No_of_Candidates_Hired",
        "State",
        "WorkExperience",
        "Country",
        "ModifiedBy",
        "Salary",
        "Remote_Job",
    ]
    tags = []
    for tag in tags_list:
        if job_opening.get(tag):
            tags.append(dict(name=tag, value=job_opening[tag]))
    return tags


def format_zoho_job_opening_to_hrflow(job_opening: dict) -> dict:
    hrflow_job = dict(
        reference=job_opening["id"],
        created_at=job_opening["Created_Time"],
        updated_at=job_opening["Modified_Time"],
        name=job_opening["Job_Opening_Name"],
        location=get_location(job_opening),
        summary=job_opening["Job_Description"],
        skills=get_skills(job_opening["Required_Skills"]),
        # TODO: get Additional_Info & Job_Description and add them only if they exist
        sections=[
            dict(
                name="Job_Description",
                title="Job Description",
                description=job_opening["Job_Description"],
            ),
            (
                dict(
                    name="Additional_Info",
                    title="Additional Info",
                    description=job_opening.get("Additional_Info", None),
                )
            ),
        ],
        tags=get_job_opening_tags(job_opening),
    )
    return hrflow_job


def get_experiences(experiences: list) -> list:
    experiences_list = []
    for experience in experiences:
        experience_dict = dict(
            title=experience["Occupation_Title"],
            company=experience["Company"],
            location={"text": "", "lng": None, "lat": None},
            date_start=experience.get("Work_Duration", {}).get("from", None),
            date_end=experience.get("Work_Duration", {}).get("to", None),
            description=experience["Summary"],
        )
        experiences_list.append(experience_dict)
    return experiences_list


def get_educations(educations: list) -> list:
    educations_list = []
    for education in educations:
        education_dict = dict(
            title=education["Degree"],
            school=education["Institute_School"],
            location={"text": "", "lng": None, "lat": None},
            date_start=education.get("Education_Duration", {}).get("from", None),
            date_end=education.get("Education_Duration", {}).get("to", None),
            description=education["Major_Department"],
        )
        educations_list.append(education_dict)
    return educations_list


def get_skills(skill_set: str) -> list:
    skills = []
    if not skill_set:
        return skills
    extracted_skills = skill_set.split(", ")
    for skill in extracted_skills:
        skill_dict = dict(
            name=skill,
            value=None,
        )
        skills.append(skill_dict)
    return skills


def get_candidate_tags(candidate: dict) -> list:
    tags_list = [
        "Current_Job_Title",
        "Current_Salary",
        "Expected_Salary",
        "Current_Employer",
        "Highest_Qualification_Held",
        "Source",
        "Origin",
        "Candidate_Status",
        "Is_Unqualified",
        "Additional_Info",
        "applied_with_linkedin",
        "No_of_Applications",
    ]
    tags = []
    for tag in tags_list:
        if candidate.get(tag):
            tags.append(dict(name=tag, value=candidate[tag]))
    return tags


def format_zoho_candidate_to_hrflow(candidate: dict) -> dict:
    hrflow_profile = dict(
        reference=candidate["id"],
        info=dict(
            first_name=candidate["First_Name"],
            last_name=candidate["Last_Name"],
            full_name=candidate["Full_Name"],
            email=candidate["Email"],
            phone=candidate["Phone"],
            location=get_location(candidate),
            urls=[
                dict(type="linkedin", url=candidate["LinkedIn__s"]),
                dict(type="facebook", url=candidate["Facebook__s"]),
                dict(type="twitter", url=candidate["Twitter"]),
                dict(type="from_resume", url=candidate["Website"]),
            ],
        ),
        created_at=candidate["Created_Time"],
        updated_at=candidate["Updated_On"],
        experiences_duration=candidate["Experience_in_Years"],
        experiences=get_experiences(candidate["Experience_Details"]),
        educations=get_educations(candidate["Educational_Details"]),
        skills=get_skills(candidate["Skill_Set"]),
        tags=get_candidate_tags(candidate),
    )
    return hrflow_profile


def get_zoho_skill_set(skills: list) -> str:
    skill_set = ""
    for skill in skills:
        skill_set += skill["name"] + ", "
    return skill_set[:-2]


def get_zoho_experiences(experiences: list) -> list:
    experiences_list = []
    for experience in experiences:
        work_duration = None
        if experience["date_start"]:
            work_duration = {
                "from": experience["date_start"],
            }
            if experience["date_end"]:
                work_duration["to"] = experience["date_end"]
        experience_dict = dict(
            Occupation_Title=experience["title"],
            I_currently_work_here=experience["date_end"] is None,
            Company=experience["company"],
            Work_Duration=work_duration,
            Summary=experience["description"],
        )
        experiences_list.append(experience_dict)
    return experiences_list


def get_zoho_educations(educations: list) -> list:
    educations_list = []
    for education in educations:
        duration = None
        if education["date_start"]:
            duration = {
                "from": education["date_start"],
            }
            if education["date_end"]:
                duration["to"] = education["date_end"]
        education_dict = dict(
            Institute_School=education["school"],
            Currently_pursuing=education["date_end"] is None,
            Degree=education["title"],
            Major_Department=education["description"],
            Duration=duration,
        )
        educations_list.append(education_dict)
    return educations_list


def get_url(urls: list, url_type: str) -> t.Optional[str]:
    for url in urls:
        if url["type"] == url_type:
            return url["url"]
    return None


def format_hrflow_profile_to_zoho(profile: dict) -> dict:
    zoho_candidate = dict(
        First_Name=profile["info"]["first_name"],
        Last_Name=profile["info"]["last_name"],
        Full_Name=profile["info"]["full_name"],
        Email=profile["info"]["email"],
        Phone=profile["info"]["phone"],
        skillSet=get_zoho_skill_set(profile["skills"]),
        Experience_in_Years=(
            int(profile["experiences_duration"])
            if profile["experiences_duration"]
            else None
        ),
        Experience_Details=get_zoho_experiences(profile["experiences"]),
        Educational_Details=get_zoho_educations(profile["educations"]),
        Created_Time=profile["created_at"][:19] if profile["created_at"] else None,
        Updated_On=profile["updated_at"][:19] if profile["updated_at"] else None,
        LinkedIn__s=get_url(profile["info"]["urls"], "linkedin"),
        Facebook__s=get_url(profile["info"]["urls"], "facebook"),
        Twitter=get_url(profile["info"]["urls"], "twitter"),
        Street=profile["info"]["location"]["text"],
        City=profile["info"]["location"]["fields"].get("city", None),
        State=profile["info"]["location"]["fields"].get("state", None),
        Zip_Code=profile["info"]["location"]["fields"].get("postcode", None),
        Country=profile["info"]["location"]["fields"].get("country", None),
    )

    return zoho_candidate


def format_hrflow_profile_for_update_to_zoho(profile: dict) -> dict:
    zoho_candidate = format_hrflow_profile_to_zoho(profile)
    zoho_candidate["id"] = int(profile["reference"])
    return zoho_candidate


def format_archive_in_hrflow(record: dict) -> dict:
    return dict(reference=record["id"])


def format_archive_in_zoho(record: dict) -> dict:
    return dict(id=record["reference"])


DESCRIPTION = (
    "Zoho Recruit offers a powerful ATS and CRM in a single recruitment platform. With"
    " scalability, customization, and remote hiring tools, Recruit has everything your"
    " staffing agency or internal HR team needs to match the right candidate to the"
    " right role."
)

ZohoRecruit = Connector(
    name="Zoho Recruit",
    type=ConnectorType.ATS,
    subtype="zohorecruit",
    description=DESCRIPTION,
    url="https://zoho.com/recruit/",
    warehouse=ZohoRecruitWarehouse,
    flows=(
        Flow(
            Mode.create,
            Entity.profile,
            Direction.inbound,
            format=format_zoho_candidate_to_hrflow,
        ),
        Flow(
            Mode.create,
            Entity.profile,
            Direction.outbound,
            format=format_hrflow_profile_to_zoho,
        ),
        Flow(
            Mode.update,
            Entity.profile,
            Direction.inbound,
            format=format_zoho_candidate_to_hrflow,
        ),
        Flow(
            Mode.update,
            Entity.profile,
            Direction.outbound,
            format=format_hrflow_profile_for_update_to_zoho,
        ),
        Flow(
            Mode.archive,
            Entity.profile,
            Direction.inbound,
            format=format_archive_in_hrflow,
        ),
        Flow(
            Mode.archive,
            Entity.profile,
            Direction.outbound,
            format=format_archive_in_zoho,
        ),
        Flow(
            Mode.create,
            Entity.job,
            Direction.inbound,
            format=format_zoho_job_opening_to_hrflow,
        ),
        Flow(
            Mode.update,
            Entity.job,
            Direction.inbound,
            format=format_zoho_job_opening_to_hrflow,
        ),
        Flow(
            Mode.archive, Entity.job, Direction.inbound, format=format_archive_in_hrflow
        ),
    ),
)



from __future__ import annotations

import logging
import typing as t
from pathlib import Path

from hrflow_connectors.v2.core import templating
from hrflow_connectors.v2.core.utils import CONNECTORS_DIRECTORY, get_import_name

if t.TYPE_CHECKING:
    from hrflow_connectors.v2.core.connector import Connector  # pragma: nocover

logger = logging.getLogger()

KEEP_EMPTY_FOLDER = ".gitkeep"


def ensure_gitkeep(directory: Path, gitkeep_filename: str = ".gitkeep") -> None:
    gitkeep_file = directory / gitkeep_filename
    create_empty_file = True

    if directory.is_dir():
        for child in directory.iterdir():
            if not child.name == gitkeep_file.name:
                create_empty_file = False
                try:
                    gitkeep_file.unlink()
                except FileNotFoundError:
                    pass
                break
    else:
        directory.mkdir(parents=True)

    if create_empty_file:
        gitkeep_file.touch()


def hrflow_connectors_docs(
    connectors: t.Iterable[Connector],
    connectors_directory: Path = CONNECTORS_DIRECTORY,
) -> None:
    for connector in connectors:
        # Done early to avoid writing file to disk if connector is
        # misconfigured
        get_import_name(connector)

        connector_directory = connectors_directory / connector.subtype
        if not connector_directory.is_dir():
            logger.error(
                "Skipping documentation for {}: no directory found at {}".format(
                    connector.name, connector_directory
                )
            )
            continue

        readme = connector_directory / "README.md"
        readme_content = None
        if readme.exists():
            readme_content = readme.read_text()

        readme.write_bytes(
            templating.connector_readme(
                connector, current_content=readme_content
            ).encode()
        )

        stubs = connector_directory / "connector.pyi"
        stubs.write_bytes(templating.connector_stub(connector).encode())

        notebooks_directory = connector_directory / "notebooks"
        ensure_gitkeep(notebooks_directory, KEEP_EMPTY_FOLDER)

        format_mappings_directory = connector_directory / "mappings" / "format"
        ensure_gitkeep(format_mappings_directory, KEEP_EMPTY_FOLDER)

        action_docs_directory = connector_directory / "docs"
        if not action_docs_directory.is_dir():
            action_docs_directory.mkdir()

        for flow in connector.flows:
            action_documentation = action_docs_directory / "{}.md".format(
                flow.name(connector.subtype)
            )
            action_documentation.write_bytes(
                templating.connector_action(connector, flow).encode()
            )



import datetime
import typing as t
from enum import Enum, IntEnum

from msgspec import Meta, Struct
from typing_extensions import Annotated

POLE_EMPLOI_REFERENCES_ENDPOINT = (
    "https://api.emploi-store.fr/partenaire/offresdemploi/v2/referentiel/"
)


def validate_date(value: t.Optional[str], field_name: str) -> None:
    if value is not None:
        try:
            datetime.date.fromisoformat(value)
        except ValueError:
            raise ValueError(
                f"Invalid date format for '{field_name}': {value}. Expected ISO 8601"
                " format (YYYY-MM-DD)."
            )


class JobLocation(Struct):
    libelle: str
    latitude: float
    longitude: float
    codepostal: str
    commune: Annotated[
        str,
        Meta(
            description=(
                "INSEE code of the commune"
                "A GET request for the list of accepted choices from the Offres"
                " d'emploi API"
                "to this endpoint :"
                f" {POLE_EMPLOI_REFERENCES_ENDPOINT}communes"
            ),
        ),
    ]


class Entreprise(Struct):
    nom: t.Optional[str]
    description: t.Optional[str]
    logo: t.Optional[str]
    url: t.Optional[str]
    entrepriseAdaptee: bool


class Partner(Struct):
    nom: str
    url: str
    logo: str


class OfferOriginTag(IntEnum):
    POLE_EMPLOI = 1
    PARTNER = 2


class OfferOrigin(Struct):
    origine: OfferOriginTag
    urlOrigine: t.Optional[str]
    partenaires: t.List[Partner]


class ExperienceRequirement(str, Enum):
    BEGINNER_ACCEPTED = "D"
    EXPERIENCE_DESIRED = "S"
    EXPERIENCE_REQUIRED = "E"


class Exigence(Enum):
    REQUIRED = "E"
    DESIRED = "S"


class Formation(Struct):
    domaineLibelle: str
    niveauLibelle: Annotated[
        str,
        Meta(
            description=(
                "Label of the level of the education required"
                "A GET request for the list of accepted choices from the Offres"
                " d'emploi API"
                "to this endpoint :"
                f" {POLE_EMPLOI_REFERENCES_ENDPOINT}referentiel/niveauxFormations"
            ),
        ),
    ]
    commentaire: str
    exigence: Exigence


class Langue(Struct):
    libelle: Annotated[
        str,
        Meta(
            description=(
                "Language label"
                "A GET request for the list of accepted choices from the Offres"
                " d'emploi API"
                "to this endpoint :"
                f" {POLE_EMPLOI_REFERENCES_ENDPOINT}referentiel/langues"
            ),
        ),
    ]
    exigence: t.Optional[Exigence]


class Permis(Struct):
    libelle: Annotated[
        str,
        Meta(
            description=(
                "requested license"
                "A GET request for the list of accepted choices from the Offres"
                " d'emploi API"
                "to this endpoint :"
                f" {POLE_EMPLOI_REFERENCES_ENDPOINT}referentiel/permis"
            ),
        ),
    ]
    exigence: t.Optional[Exigence]


class Competence(Struct):
    code: str
    libelle: str
    exigence: t.Optional[Exigence]


class Salaire(Struct):
    libelle: t.Optional[str]
    commentaire: t.Optional[str]
    complement1: t.Optional[str]
    complement2: t.Optional[str]


class Contact(Struct):
    nom: t.Optional[str]
    coordonnees1: t.Optional[str]
    coordonnees2: t.Optional[str]
    coordonnees3: t.Optional[str]
    telephone: t.Optional[str]
    courriel: t.Optional[str]
    commentaire: t.Optional[str]
    urlRecruteur: t.Optional[str]
    urlPostulation: t.Optional[str]


class Agence(Struct):
    telephone: t.Optional[str]
    courriel: t.Optional[str]


class QualificationCode(str, Enum):
    MANOEUVRE = "1"
    SPECIALIZED_WORKER = "2"
    QUALIFIED_WORKER_P1_P2 = "3"
    QUALIFIED_WORKER_P3_P4_OHQ = "4"
    UNQUALIFIED_EMPLOYEE = "5"
    QUALIFIED_EMPLOYEE = "6"
    TECHNICIAN = "7"
    SUPERVISORY_AGENT = "8"
    EXECUTIVE = "9"


class QualificationLibelle(str, Enum):
    MANOEUVRE = "Manœuvre"
    SPECIALIZED_WORKER = "Ouvrier spécialisé"
    QUALIFIED_WORKER_P1_P2 = "Ouvrier qualifié (P1, P2)"
    QUALIFIED_WORKER_P3_P4_OHQ = "Ouvrier qualifié (P3, P4, OHQ)"
    UNQUALIFIED_EMPLOYEE = "Employé non qualifié"
    QUALIFIED_EMPLOYEE = "Employé qualifié"
    TECHNICIAN = "Technicien"
    SUPERVISORY_AGENT = "Agent de maîtrise"
    EXECUTIVE = "Cadre"


class QualitePro(Struct):
    libelle: t.Optional[str]
    description: t.Optional[str]


class FranceTravailJobOffer(Struct):
    id: int
    intitule: str
    description: str
    dateCreation: t.Optional[str] = None
    dateActualisation: t.Optional[str] = None
    lieuTravail: t.Optional[JobLocation] = None
    romeCode: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "ROME code of the profession"
                "A GET request for the list of accepted choices from the Offres"
                " d'emploi API"
                "to this endpoint :"
                f" {POLE_EMPLOI_REFERENCES_ENDPOINT}metiers"
            ),
        ),
    ] = None
    romeLibelle: t.Optional[str] = None
    appellationLibelle: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "Code of the appellation"
                "A GET request for the list of accepted choices from the Offres"
                " d'emploi API"
                "to this endpoint :"
                f" {POLE_EMPLOI_REFERENCES_ENDPOINT}appellations"
            ),
        ),
    ] = None
    entreprise: t.Optional[Entreprise] = None
    typeContrat: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "Contract type code"
                "A GET request for the list of accepted choices from the Offres"
                " d'emploi API"
                "to this endpoint :"
                f" {POLE_EMPLOI_REFERENCES_ENDPOINT}typesContrats"
            ),
        ),
    ] = None
    typeContratLibelle: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "Contract type label"
                "Example : CDI,CDD"
                "A GET request for the list of accepted choices from the Offres"
                " d'emploi API"
                "to this endpoint :"
                f" {POLE_EMPLOI_REFERENCES_ENDPOINT}typesContrats"
            ),
        ),
    ] = None
    natureContrat: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "Code of the nature of contract"
                "A GET request for the list of accepted choices from the Offres"
                " d'emploi API"
                "to this endpoint :"
                f" {POLE_EMPLOI_REFERENCES_ENDPOINT}naturesContrats"
            ),
        ),
    ] = None
    origineOffre: t.Optional[OfferOrigin] = None
    offresManqueCandidats: t.Optional[bool] = None
    experienceExige: t.Optional[ExperienceRequirement] = None
    experienceLibelle: t.Optional[str] = None
    experienceCommentaire: t.Optional[str] = None
    formations: t.Optional[t.List[Formation]] = None
    langues: t.Optional[t.List[Langue]] = None
    permis: t.Optional[t.List[Permis]] = None
    outilsBureautiques: t.Optional[str] = None
    competences: t.Optional[t.List[Competence]] = None
    salaire: t.Optional[Salaire] = None
    dureeTravailLibelle: t.Optional[str] = None
    dureeTravailLibelleConverti: t.Optional[str] = None
    complementExercice: t.Optional[str] = None
    conditionExercice: t.Optional[str] = None
    alternance: t.Optional[bool] = None
    contact: t.Optional[Contact] = None
    agence: t.Optional[Agence] = None
    nombrePostes: t.Optional[int] = None
    accessibleTH: t.Optional[bool] = None
    deplacementCode: t.Optional[str] = None
    deplacementLibelle: t.Optional[str] = None
    qualificationCode: t.Optional[QualificationCode] = None
    qualificationLibelle: t.Optional[QualificationLibelle] = None
    secteurActivite: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "NAF codes for sectors of activity. It is possible to specify two NAF"
                " codes by separating them with a comma in the character"
                " string.Example : 01,02A GET request for the list of accepted choices"
                " from the Offres d'emploi APIto this endpoint :"
                f" {POLE_EMPLOI_REFERENCES_ENDPOINT}secteursActivites"
            ),
        ),
    ] = None
    secteurActiviteLibelle: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "Sector of activitylabel"
                "A GET request for the list of accepted choices from the Offres"
                " d'emploi API"
                "to this endpoint :"
                f" {POLE_EMPLOI_REFERENCES_ENDPOINT}secteursActivites"
            ),
        ),
    ] = None
    qualitesProfessionnelles: t.Optional[t.List[QualitePro]] = None

    def __post_init__(self):
        # Validate date fields
        validate_date(self.dateCreation, "dateCreation")
        validate_date(self.dateActualisation, "dateActualisation")



import enum
import pdb
import typing as t
from logging import LoggerAdapter

import requests
from msgspec import Meta, Struct
from msgspec.structs import asdict
from typing_extensions import Annotated

from hrflow_connectors.v2.connectors.teamtailor.schema import (
    TeamtailorCandidateAttributes,
    TeamtailorJob,
)
from hrflow_connectors.v2.core.common import Entity
from hrflow_connectors.v2.core.warehouse import (
    Aisle,
    Criterias,
    ReadOperation,
    WriteOperation,
    merge,
)

JOBS_ENDPOINT = "https://api.teamtailor.com/v1/jobs"
CANDIDATES_ENDPOINT = "https://api.teamtailor.com/v1/candidates"


class status(str, enum.Enum):
    published = "published"
    unlisted = "unlisted"
    archived = "archived"
    draft = "draft"
    scheduled = "scheduled"
    all = "all"


class RemoteStatus(str, enum.Enum):
    none = "none"
    hybrid = "hybrid"
    temporary = "temporary"
    fully = "fully"


class JobFeed(str, enum.Enum):
    public = "public"
    internal = "internal"


class TeamTailorBoolean(str, enum.Enum):
    true = "true"
    false = "false"


class AuthParameters(Struct):
    api_key: Annotated[
        str,
        Meta(
            description=(
                "API key for authenticating with the Teamtailor API. You can generate"
                " it in the Teamtailor app under Settings > Integrations > API Keys."
            ),
        ),
    ]

    X_Api_Version: Annotated[
        str,
        Meta(
            description="API version for the Teamtailor API. Default is '20240404'.",
        ),
    ] = "20240404"


class WriteProfilesParameters(Struct):
    pass


class ReadJobsParameters(Struct):
    filter_status: Annotated[
        t.Optional[status],
        Meta(
            description="Filter by job status. Available statuses: {}".format(
                [e.value for e in status]
            ),
        ),
    ] = None

    filter_feed: Annotated[
        t.Optional[JobFeed],
        Meta(
            description="Status of a job. One of {}".format([e.value for e in JobFeed]),
        ),
    ] = None

    filter_department: Annotated[
        t.Optional[str],
        Meta(description="Filter by department id"),
    ] = None

    filter_role: Annotated[
        t.Optional[str],
        Meta(description="Filter by role id"),
    ] = None

    filter_locations: Annotated[
        t.Optional[str],
        Meta(description="Filter by location id"),
    ] = None

    filter_regions: Annotated[
        t.Optional[str],
        Meta(description="Filter by region id"),
    ] = None

    filter_tags: Annotated[
        t.Optional[str],
        Meta(description="Filter by tags"),
    ] = None

    filter_remote_status: Annotated[
        t.Optional[RemoteStatus],
        Meta(
            description=(
                "Filter by remote status. Available remote statuses: {}".format(
                    [e.value for e in RemoteStatus]
                )
            ),
        ),
    ] = None

    filter_created_at_from: Annotated[
        t.Optional[str],
        Meta(description="Filter by created-at older than this date."),
    ] = None

    filter_created_at_to: Annotated[
        t.Optional[str],
        Meta(description="Filter by created-at newer than this date."),
    ] = None

    filter_updated_at_from: Annotated[
        t.Optional[str],
        Meta(description="Filter by updated-at older than this date."),
    ] = None

    filter_updated_at_to: Annotated[
        t.Optional[str],
        Meta(description="Filter by updated-at newer than this date."),
    ] = None

    sort: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "Default sorting based on pinned status and publish"
                " date.\n'-pinned,date'"
            )
        ),
    ] = None


class ReadProfilesParameters(Struct):
    filter_email: Annotated[
        t.Optional[str],
        Meta(description="Filter by email address"),
    ] = None

    filter_department: Annotated[
        t.Optional[str],
        Meta(description="Filter by department"),
    ] = None

    filter_role: Annotated[
        t.Optional[str],
        Meta(description="Filter by role"),
    ] = None

    filter_locations: Annotated[
        t.Optional[str],
        Meta(description="Filter by location"),
    ] = None

    filter_regions: Annotated[
        t.Optional[str],
        Meta(description="Filter by region"),
    ] = None

    filter_created_at_from: Annotated[
        t.Optional[str],
        Meta(description="Filter by created-at older than this date."),
    ] = None

    filter_created_at_to: Annotated[
        t.Optional[str],
        Meta(description="Filter by created-at newer than this date."),
    ] = None

    filter_updated_at_from: Annotated[
        t.Optional[str],
        Meta(description="Filter by updated-at older than this date."),
    ] = None

    filter_updated_at_to: Annotated[
        t.Optional[str],
        Meta(description="Filter by updated-at newer than this date."),
    ] = None

    filter_connected: Annotated[
        t.Optional[TeamTailorBoolean],
        Meta(description="Filter candidates who has connected."),
    ] = None

    sort: Annotated[
        t.Optional[str],
        Meta(
            description=(
                "Sort list by any of the candidate attributes. Use -id to sort by id"
                " descending."
            ),
        ),
    ] = None


def enrich_location(job_id: str, headers: t.Dict) -> t.Dict:
    """
    Get_location sends a request to get the job location from its API endpoint
    """
    response = requests.get(
        "{}/{}/location".format(JOBS_ENDPOINT, job_id),
        headers=headers,
    )

    if not response.ok:
        raise Exception("Failed to get job location from Teamtailor.")

    location = response.json().get("data")
    if location is not None:
        location_attribute = location.get("attributes")
        text = location_attribute["address"]
        city = location_attribute["city"]
        country = location_attribute["country"]
        lat = float(location_attribute["lat"])
        lng = float(location_attribute["long"])
        zip = location_attribute["zip"]
        geojson = dict(
            text=text,
            city=city,
            country=country,
            zip=zip,
        )
        location_obj = dict(text=text, geojson=geojson, lat=lat, lng=lng)
        return location_obj
    return dict(text="", geojson="", lat=None, lng=None)


def read_jobs(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
    parameters: ReadJobsParameters,
    incremental: bool,
    incremental_token: t.Optional[str],
) -> t.Iterable[t.Dict]:
    headers = {
        "Authorization": f"Token token={auth_parameters.api_key}",
        "X-Api-Version": auth_parameters.X_Api_Version,
    }

    params = asdict(parameters)
    params_dict = {
        "filter[status]": params.get("filter_status"),
        "filter[feed]": params.get("filter_feed"),
        "filter[department]": params.get("filter_department"),
        "filter[role]": params.get("filter_role"),
        "filter[locations]": params.get("filter_locations"),
        "filter[regions]": params.get("filter_regions"),
        "filter[tags]": params.get("filter_tags"),
        "filter[remote-status]": params.get("filter_remote_status"),
        "filter[created-at][from]": params.get("filter_created_at_from"),
        "filter[created-at][to]": params.get("filter_created_at_to"),
        "filter[updated-at][from]": params.get("filter_updated_at_from"),
        "filter[updated-at][to]": params.get("filter_updated_at_to"),
        "sort": params.get("sort"),
    }

    params_dict = {k: v for k, v in params_dict.items() if v is not None}

    url = JOBS_ENDPOINT
    all_jobs = []
    while True:
        response = requests.get(url, headers=headers, params=params_dict)
        if response.status_code // 100 != 2:
            adapter.error(
                "Failed to pull jobs from Teamtailor status_code={} response={}".format(
                    response.status_code, response.text
                )
            )
            raise Exception("Failed to pull jobs from Teamtailor")
        response = response.json()
        jobs = response.get("data")

        all_jobs += jobs
        if response.get("links").get("next") is None:
            break
        else:
            url = response.get("links").get("next")

    adapter.info("Pulling {} jobs from Teamtailor API".format(len(all_jobs)))

    for job in all_jobs:
        job_location = enrich_location(job.get("id"), headers)
        job["location"] = job_location
        yield job


def read_profiles(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
    parameters: ReadProfilesParameters,
    incremental: bool,
    incremental_token: t.Optional[str],
) -> t.Iterable[t.Dict]:
    headers = {
        "Authorization": f"Token token={auth_parameters.api_key}",
        "X-Api-Version": auth_parameters.X_Api_Version,
    }

    params = asdict(parameters)
    params_dict = {
        "filter[email]": params.get("filter_email"),
        "filter[department]": params.get("filter_department"),
        "filter[role]": params.get("filter_role"),
        "filter[locations]": params.get("filter_locations"),
        "filter[regions]": params.get("filter_regions"),
        "filter[created-at][from]": params.get("filter_created_at_from"),
        "filter[created-at][to]": params.get("filter_created_at_to"),
        "filter[updated-at][from]": params.get("filter_updated_at_from"),
        "filter[updated-at][to]": params.get("filter_updated_at_to"),
        "filter[connected]": params.get("filter_connected"),
        "sort": params.get("sort"),
    }

    params_dict = {k: v for k, v in params_dict.items() if v is not None}

    url = CANDIDATES_ENDPOINT
    all_profiles = []
    while True:
        response = requests.get(url, headers=headers, params=params_dict)
        if response.status_code // 100 != 2:
            adapter.error(
                "Failed to pull profiles from Teamtailor status_code={} response={}"
                .format(response.status_code, response.text)
            )
            raise Exception("Failed to pull profiles from Teamtailor")
        response = response.json()
        profiles = response.get("data")

        all_profiles += profiles
        if response.get("links").get("next") is None:
            break
        else:
            url = response.get("links").get("next")

    adapter.info("Pulling {} profiles from Teamtailor API".format(len(all_profiles)))

    for profile in all_profiles:
        resume_url = profile["attributes"].get("resume")
        if resume_url is not None:
            resume_response = requests.get(resume_url)
            if resume_response.status_code // 100 != 2:
                adapter.error(
                    "Failed to get resume from Teamtailor status_code={} response={}"
                    .format(resume_response.status_code, resume_response.text)
                )
                raise Exception("Failed to get resume from Teamtailor")
            profile["resume"] = dict(raw=resume_response.content)
        yield profile


def write_profiles(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
    parameters: WriteProfilesParameters,
    items: t.Iterable[t.Dict],
) -> t.List[t.Dict]:
    failed_profiles = []

    for profile in items:
        pdb.set_trace()
        profile_obj = {
            "data": {
                "type": "candidates",
                "attributes": profile,
            },
        }

        headers = {
            "Authorization": f"Token token={auth_parameters.api_key}",
            "X-Api-Version": auth_parameters.X_Api_Version,
            "Content-Type": "application/vnd.api+json",
        }

        response = requests.post(CANDIDATES_ENDPOINT, headers=headers, json=profile_obj)

        if response.status_code // 100 != 2:
            adapter.error(
                "Failed to push profile to Teamtailor,  status_code={} response={}"
                .format(
                    response.status_code,
                    response.text,
                )
            )
            failed_profiles.append(profile)
    return failed_profiles


def update_profiles(
    adapter: LoggerAdapter,
    auth_parameters: AuthParameters,
    parameters: WriteProfilesParameters,
    items: t.Iterable[t.Dict],
) -> t.List[t.Dict]:
    failed_profiles = []

    for profile in items:
        profile_id = profile.pop("id")
        profile_obj = {
            "data": {
                "id": profile_id,
                "attributes": profile,
                "type": "candidates",
            },
        }

        headers = {
            "Authorization": f"Token token={auth_parameters.api_key}",
            "X-Api-Version": auth_parameters.X_Api_Version,
            "Content-Type": "application/vnd.api+json",
        }
        response = requests.patch(
            "{}/{}".format(CANDIDATES_ENDPOINT, profile_id),
            headers=headers,
            json=profile_obj,
        )

        if response.status_code // 100 != 2:
            adapter.error(
                "Failed to update profile to Teamtailor,  status_code={} response={}"
                .format(
                    response.status_code,
                    response.text,
                )
            )
            failed_profiles.append(profile)
    return failed_profiles


JobsAisle = Aisle(
    name=Entity.job,
    schema=TeamtailorJob,
    read=ReadOperation(
        criterias=Criterias(
            create=ReadJobsParameters,
            update=ReadJobsParameters,
            archive=ReadJobsParameters,
        ),
        function=merge(
            create=read_jobs,
            update=read_jobs,
            archive=read_jobs,
        ),
    ),
)

ProfilesAisle = Aisle(
    name=Entity.profile,
    schema=TeamtailorCandidateAttributes,
    read=ReadOperation(
        criterias=Criterias(
            create=ReadProfilesParameters,
            update=ReadProfilesParameters,
            archive=ReadProfilesParameters,
        ),
        function=merge(
            create=read_profiles,
            update=read_profiles,
            archive=read_profiles,
        ),
    ),
    write=WriteOperation(
        criterias=Criterias(
            create=WriteProfilesParameters,
            update=WriteProfilesParameters,
        ),
        function=merge(
            create=write_profiles,
            update=update_profiles,
        ),
    ),
)



# This file is generated automatically
from hrflow_connectors.v2.core.connector import Connector, PublicActionInterface

class AdmenProto(Connector):
    create_jobs_in_hrflow: PublicActionInterface
    update_jobs_in_hrflow: PublicActionInterface
    archive_jobs_in_hrflow: PublicActionInterface
    create_profiles_in_hrflow: PublicActionInterface
    update_profiles_in_hrflow: PublicActionInterface
    archive_profiles_in_hrflow: PublicActionInterface
    create_profiles_in_admen: PublicActionInterface
    update_profiles_in_admen: PublicActionInterface
    archive_profiles_in_admen: PublicActionInterface
    
Admen: AdmenProto


from hrflow_connectors.v2.connectors.waalaxy.aisles import AuthParameters, ProfilesAisle
from hrflow_connectors.v2.core.warehouse import Warehouse

WaalaxyWarehouse = Warehouse(auth=AuthParameters, aisles=(ProfilesAisle,))



from typing import Any, Dict, List, Optional

from msgspec import Meta, Struct
from typing_extensions import Annotated


class Type(Struct):
    id: str
    name: str


class Experience(Struct):
    id: str
    name: str


class Country(Struct):
    name: str
    id: str


class Location(Struct):
    country: Country
    city: str
    is_remote: Optional[bool]
    name: str


class Category(Struct):
    id: str
    name: str


class ApplicationForm(Struct):
    name: str
    headline: str
    summary: str
    profile_photo: str
    address: str
    email_address: str
    phone_number: str
    resume: str
    work_history: str
    education: str
    cover_letter: str
    questionnaire_in_experience: bool


class BreezyJobModel(Struct):
    _id: Annotated[str, Meta(description="position id")]
    type: Annotated[Type, Meta(description="job type")]
    state: Annotated[
        str,
        Meta(
            description="state of the position posting, published or internal and so on"
        ),
    ]
    name: Annotated[str, Meta(description="job title")]
    friendly_id: Annotated[
        str,
        Meta(description="another id of the job which combines its title and its id"),
    ]
    experience: Optional[Experience]
    location: Location
    education: str
    department: str
    description: Annotated[str, Meta(description="Job category")]
    category: Category
    application_form: Annotated[
        Optional[ApplicationForm], Meta(description="job Application for")
    ]
    creator_id: Optional[str]
    creation_date: str
    updated_date: str
    all_users: List[str]
    all_admins: List[str]
    candidate_type: str
    tags: List
    org_type: str


class WorkHistoryItem(Struct):
    company_name: str
    title: str
    summary: str
    start_month: Optional[int]
    start_year: Optional[int]
    end_month: Optional[int]
    end_year: Optional[int]


class EducationItem(Struct):
    school_name: str
    field_of_study: str
    start_year: Optional[int]
    end_year: Optional[int]


class BreezyProfileModel(Struct):
    name: str
    email_address: str
    phone_number: str
    summary: str
    tags: Optional[List[str]]
    source: Optional[str]
    origin: Optional[str]
    address: str
    work_history: List[WorkHistoryItem]
    education: List[EducationItem]
    social_profiles: Optional[List[str]]
    custom_attributes: Optional[List[Dict[str, Any]]]
    cover_letter: Optional[str]



from msgspec import Struct


class WaalaxyProfile(Struct):
    _id: str
    firstName: str
    lastName: str
    occupation: str
    location: str
    proEmail: str
    phoneNumbers: str
    prospectList: str
    linkedinUrl: str
    premium: str
    jobSeeker: str
    profileStatus: str
    messageSent: str
    messageReplied: str
    emailSent: str
    emailReplied: str
    connectedAt: str
    firstMessageAt: str
    tags: str
    company_name: str
    company_linkedinUrl: str
    company_website: str
    job_title: str
    linkedinEmail: str
    salesNavigatorId: str
    profileId: str



from hrflow_connectors.v1.connectors.hrflow.warehouse.job import (  # noqa
    HrFlowJobWarehouse,
)
from hrflow_connectors.v1.connectors.hrflow.warehouse.profile import (  # noqa
    HrFlowProfileParsingWarehouse,
    HrFlowProfileWarehouse,
)



import typing as t

from hrflow_connectors.v2.connectors.hubspot.warehouse import HubspotWarehouse
from hrflow_connectors.v2.core.common import Direction, Entity, Mode
from hrflow_connectors.v2.core.connector import Connector, ConnectorType, Flow


def format_hrflow_profile(hrflow_profile: t.Dict) -> t.Dict:
    hrflow_profile_info = hrflow_profile["info"]
    hrflow_location = hrflow_profile_info["location"]
    fields = hrflow_location["fields"] or {}
    properties = dict(
        email=hrflow_profile_info["email"],
        firstname=hrflow_profile_info["first_name"],
        lastname=hrflow_profile_info["last_name"],
        date_of_birth=hrflow_profile_info["date_birth"],
        phone=hrflow_profile_info["phone"],
        address=hrflow_location["text"],
        zip=fields.get("postcode") or "Undefined",
        city=fields.get("city") or "Undefined",
        state=fields.get("state") or "Undefined",
        country=fields.get("country") or "Undefined",
        company="",
    )
    hubspot_contact = dict(properties=properties)
    return hubspot_contact


def format_hrflow_profile_for_update(hrflow_profile: t.Dict) -> t.Dict:
    hubspot_contact = format_hrflow_profile(hrflow_profile)
    hubspot_contact["contactId"] = hrflow_profile["reference"]
    return hubspot_contact


def format_hrflow_profile_for_archive(hrflow_profile: t.Dict) -> t.Dict:
    hubspot_contact = dict(contactId=hrflow_profile["reference"])
    return hubspot_contact


def format_hubspot_contact(hubspot_contact: t.Dict) -> t.Dict:
    properties = hubspot_contact["properties"]
    profile = dict(
        reference=hubspot_contact["id"],
        info=dict(
            email=properties["email"],
            first_name=properties["firstname"],
            last_name=properties["lastname"],
            full_name=f"{properties['firstname']} {properties['lastname']}",
            date_birth=properties["date_of_birth"],
            phone=properties["phone"],
            location=dict(
                text=properties["address"],
                fields=dict(
                    postcode=properties["zip"],
                    city=properties["city"],
                    state=properties["state"],
                    country=properties["country"],
                ),
            ),
        ),
        created_at=hubspot_contact["createdAt"],
        updated_at=hubspot_contact["updatedAt"],
        experiences=[],
        educations=[],
    )
    return profile


def format_hubspot_contact_for_archive(hubspot_contact: t.Dict) -> t.Dict:
    return dict(reference=hubspot_contact["id"])


Hubspot = Connector(
    name="Hubspot",
    type=ConnectorType.CRM,
    subtype="hubspot",
    description=(
        "HubSpot is a CRM platform with all the software, integrations, and resources"
        " you need to connect marketing, sales, content management, and customer"
        " service."
    ),
    url="https://www.hubspot.com/",
    warehouse=HubspotWarehouse,
    flows=(
        Flow(
            Mode.create,
            Entity.profile,
            Direction.inbound,
            format=format_hubspot_contact,
        ),
        Flow(
            Mode.create,
            Entity.profile,
            Direction.outbound,
            format=format_hrflow_profile,
        ),
        Flow(
            Mode.update,
            Entity.profile,
            Direction.inbound,
            format=format_hubspot_contact,
        ),
        Flow(
            Mode.update,
            Entity.profile,
            Direction.outbound,
            format=format_hrflow_profile_for_update,
        ),
        Flow(
            Mode.archive,
            Entity.profile,
            Direction.inbound,
            format=format_hubspot_contact_for_archive,
        ),
        Flow(
            Mode.archive,
            Entity.profile,
            Direction.outbound,
            format=format_hrflow_profile_for_archive,
        ),
    ),
)



import typing as t

from hrflow_connectors.v2.connectors.admen.warehouse import AdmenWarehouse
from hrflow_connectors.v2.core.common import Direction, Entity, Mode
from hrflow_connectors.v2.core.connector import Connector, ConnectorType, Flow


def get_job_location(admen_job: t.Dict) -> t.Dict:
    adress_components = []
    for key in ["CITY", "ZIP", "REGION", "PAYS"]:
        if admen_job[key]:
            adress_components.append(admen_job[key])
    return dict(
        text=", ".join(adress_components),
        lat=None,
        lng=None,
        fields=dict(
            city=admen_job["CITY"],
            country=admen_job["PAYS"],
            postcode=admen_job["ZIP"],
            state=admen_job["REGION"],
        ),
    )


def get_job_sections(
    admen_job: t.Dict,
) -> t.List[t.Dict]:
    sections = []
    for section_name in [
        "DESC_MISSION",
        "DESC_CUSTOMER",
        "HTML",
        "POSTIT",
        "COMMENTAIRE",
    ]:
        section = admen_job[section_name]
        if section:
            sections.append(
                dict(name=section_name, title=section_name, description=section)
            )
    return sections


def get_job_tags(admen_job: t.Dict) -> t.List[str]:
    tags = []

    if admen_job["DATE_DEB_RECHERCHE"]:
        tags.append(
            {
                "name": "DATE_DEB_RECHERCHE",
                "value": admen_job["DATE_DEB_RECHERCHE"].strftime("%Y-%m-%d"),
            }
        )
    if admen_job["DATE_PROPOSITION"]:
        tags.append(
            {
                "name": "DATE_PROPOSITION",
                "value": admen_job["DATE_PROPOSITION"].strftime("%Y-%m-%d"),
            }
        )
    if admen_job["DATE_ENTREE_FONCT"]:
        tags.append(
            {
                "name": "DATE_ENTREE_FONCT",
                "value": admen_job["DATE_ENTREE_FONCT"].strftime("%Y-%m-%d"),
            }
        )
    if admen_job["SECTEUR"]:
        tags.append({"name": "SECTEUR", "value": admen_job["SECTEUR"]})
    if admen_job["FONCTION"]:
        tags.append({"name": "FONCTION", "value": admen_job["FONCTION"]})
    if admen_job["TYPEJOB"]:
        tags.append({"name": "TYPEJOB", "value": admen_job["TYPEJOB"]})
    if admen_job["TYPECONTRAT"]:
        tags.append({"name": "TYPECONTRAT", "value": admen_job["TYPECONTRAT"]})
    if admen_job["EDUCLEVEL"]:
        tags.append({"name": "EDUCLEVEL", "value": admen_job["EDUCLEVEL"]})
    if admen_job["JOBEXP"]:
        tags.append({"name": "JOBEXP", "value": admen_job["JOBEXP"]})
    if admen_job["PERIOD"]:
        tags.append({"name": "PERIOD", "value": admen_job["PERIOD"]})
    if admen_job["SALARYMIN"]:
        tags.append({"name": "SALARYMIN", "value": str(admen_job["SALARYMIN"])})
    if admen_job["SALARYMAX"]:
        tags.append({"name": "SALARYMAX", "value": str(admen_job["SALARYMAX"])})
    if admen_job["SALAIRE_ENTREE"]:
        tags.append(
            {"name": "SALAIRE_ENTREE", "value": str(admen_job["SALAIRE_ENTREE"])}
        )
    if admen_job["PROCESS"]:
        tags.append({"name": "PROCESS", "value": admen_job["PROCESS"]})
    return tags


def format_job(
    admen_job: t.Dict,
) -> t.Dict:
    job = dict(
        name=admen_job["LIBELLE"],
        reference=str(admen_job["ID_MISSION"]),
        created_at=admen_job["DATE_SAISIE"].strftime("%Y-%m-%d"),
        updated_at=admen_job["DATE_MAJ"].strftime("%Y-%m-%d"),
        location=get_job_location(admen_job),
        url=admen_job["URL"],
        summary=admen_job["DESC_MISSION"],
        sections=get_job_sections(admen_job),
        tags=get_job_tags(admen_job),
    )
    return job


def get_url(type: str, urls: t.List[t.Dict]) -> t.Optional[str]:
    return next(
        (url["url"] for url in urls if url["type"] == type),
        None,
    )


def format_hrflow_experiences_to_admen(hrflow_experiences: t.List) -> t.List[t.Dict]:
    experiences = []
    for hrflow_experience in hrflow_experiences:
        experience = dict(
            INTITULE_POSTE=hrflow_experience.get("title"),
            RAISON_SOCIALE=hrflow_experience.get("company"),
            POSTE_OCCUPE=hrflow_experience.get("title"),
            DEPARTMENT=hrflow_experience["location"].get("text"),
            DATE_EXP=hrflow_experience["date_start"][:10],
            DATE_FIN=(
                hrflow_experience.get("date_end")[:10]
                if hrflow_experience.get("date_end")
                else None
            ),
            COMMENTAIRES=hrflow_experience.get("description"),
            EN_COURS=1 if not hrflow_experience.get("date_end") else 0,
        )
        experiences.append(experience)
    return experiences


def format_hrflow_profile_to_admen(
    hrflow_profile: t.Dict,
) -> t.Dict:
    hrflow_profile_info = hrflow_profile["info"]

    admen_profile = dict(
        NOM=hrflow_profile_info["last_name"],
        PRENOM=hrflow_profile_info["first_name"],
        EMAIL_PERSO=hrflow_profile_info["email"],
        TEL_MOBILE=hrflow_profile_info["phone"],
        DATE_NAISSANCE=hrflow_profile_info["date_birth"],
        DATE_CREATION=hrflow_profile["created_at"][:10],
        DATE_MAJ=hrflow_profile["updated_at"],
        ADRESSE=hrflow_profile_info["location"]["text"],
        VILLE=hrflow_profile_info["location"]["fields"]["city"],
        CODE_POSTAL=hrflow_profile_info["location"]["fields"]["postcode"],
        AREA=hrflow_profile_info["location"]["fields"]["state"],
        PAYS=hrflow_profile_info["location"]["fields"]["country"],
        LINKEDINPUBLICPROFIL=get_url("linkedin", hrflow_profile_info["urls"]),
        VIADEOPUBLICPROFIL=get_url("viadeo", hrflow_profile_info["urls"]),
        TWITTERPAGE=get_url("twitter", hrflow_profile_info["urls"]),
        FACEBOOKPAGE=get_url("facebook", hrflow_profile_info["urls"]),
        experiences=format_hrflow_experiences_to_admen(hrflow_profile["experiences"]),
        scsync_modified=0,
    )
    return admen_profile


def get_profile_location(admen_profile: t.Dict) -> t.Dict:
    adress_components = []
    for key in ["ADRESSE", "CODE_POSTAL", "VILLE", "AREA", "PAYS"]:
        if admen_profile[key]:
            adress_components.append(admen_profile[key])
    return dict(
        text=admen_profile["ADRESSE"],
        lat=None,
        lng=None,
        fields=dict(
            city=admen_profile["VILLE"],
            country=admen_profile["PAYS"],
            postcode=admen_profile["CODE_POSTAL"],
            state=admen_profile["AREA"],
            text=",".join(adress_components),
        ),
    )


def get_profile_urls(admen_profile: t.Dict) -> t.List[t.Dict]:
    urls = []
    if admen_profile["LINKEDINPUBLICPROFIL"]:
        urls.append(dict(url=admen_profile["LINKEDINPUBLICPROFIL"], type="linkedin"))
    if admen_profile["FACEBOOKPAGE"]:
        urls.append(dict(url=admen_profile["FACEBOOKPAGE"], type="facebook"))
    if admen_profile["TWITTERPAGE"]:
        urls.append(dict(url=admen_profile["TWITTERPAGE"], type="twitter"))
    return urls


def format_profile_gender(admen_gender: t.Optional[str]) -> t.Optional[str]:
    if admen_gender == "F":
        return "female"
    if admen_gender == "M":
        return "male"
    return None


def get_profile_experiences(admen_profile_experiences: t.List) -> t.List[t.Dict]:
    experiences = []

    for admen_experience in admen_profile_experiences:
        if admen_experience["DATE_EXP"]:
            date_start = admen_experience["DATE_EXP"].strftime("%Y-%m-%d")
        else:
            date_start = None
        if admen_experience["DATE_FIN"]:
            date_end = admen_experience["DATE_FIN"].strftime("%Y-%m-%d")
        else:
            date_end = None
        experience = dict(
            title=admen_experience.get("INTITULE_POSTE"),
            company=admen_experience.get("RAISON_SOCIALE"),
            location=dict(
                text=admen_experience.get("VILLE"),
                lat=None,
                lng=None,
            ),
            date_start=date_start,
            date_end=date_end,
            description=admen_experience.get("POSTE_OCCUPE"),
        )
        experiences.append(experience)
    return experiences


def get_profile_tags(admen_profile: t.Dict) -> t.List[str]:
    tags = []
    if admen_profile["NATIONALITE"]:
        tags.append({"name": "NATIONALITE", "value": admen_profile["NATIONALITE"]})
    if admen_profile["SITU_FAMILLE"]:
        tags.append(
            {"name": "SITUATION_FAMILLE", "value": admen_profile["SITU_FAMILLE"]}
        )
    if admen_profile["POSTE_OCCUPE"]:
        tags.append({"name": "POSTE_OCCUPE", "value": admen_profile["POSTE_OCCUPE"]})
    if admen_profile["POSTE_RECHERCHE"]:
        tags.append(
            {"name": "POSTE_RECHERCHE", "value": admen_profile["POSTE_RECHERCHE"]}
        )
    if admen_profile["MOBILITE_TXT"]:
        tags.append({"name": "MOBILITE", "value": admen_profile["MOBILITE_TXT"]})
    if admen_profile["SECTEUR"]:
        tags.append({"name": "SECTEUR", "value": admen_profile["SECTEUR"]})
    return tags


def format_admen_profile_to_hrflow(admen_profile: t.Dict) -> t.Dict:
    admen_profile_experiences = admen_profile["experiences"]
    phone_number = (
        admen_profile["TEL_MOBILE"]
        or admen_profile["MOBILE_PRO"]
        or admen_profile["TEL_STANDARD"]
        or admen_profile["TEL_PERSO"]
    )
    email = admen_profile["EMAIL_PERSO"] or admen_profile["EMAIL"]
    if admen_profile["DATE_NAISSANCE"]:
        date_birth = admen_profile["DATE_NAISSANCE"].strftime("%Y-%m-%d")
    else:
        date_birth = None
    hrflow_profile_info = dict(
        first_name=admen_profile["PRENOM"],
        last_name=admen_profile["NOM"],
        full_name=" ".join([admen_profile["PRENOM"], admen_profile["NOM"]]),
        email=email,
        phone=phone_number,
        date_birth=date_birth,
        location=get_profile_location(admen_profile),
        urls=get_profile_urls(admen_profile),
        gender=format_profile_gender(admen_profile["GENDER"]),
    )
    hrflow_profile = dict(
        reference=str(admen_profile["ID_PERSONNE"]),
        info=hrflow_profile_info,
        resume=admen_profile["resume"],
        experiences=get_profile_experiences(admen_profile_experiences),
        educations=[],
        attachments=[],
        tags=get_profile_tags(admen_profile),
    )
    return hrflow_profile


def format_job_archive_in_hrflow(admen_job: t.Dict) -> t.Dict:
    return dict(reference=str(admen_job["ID_MISSION"]))


def format_profile_archive_in_hrflow(admen_profile: t.Dict) -> t.Dict:
    return dict(reference=str(admen_profile["ID_PERSONNE"]))


def format_hrflow_profile_for_update_to_admen(hrflow_profile: t.Dict) -> t.Dict:
    admen_profile = format_hrflow_profile_to_admen(hrflow_profile)
    admen_profile["ID_PERSONNE"] = int(hrflow_profile["reference"])
    return admen_profile


def format_profile_archive_in_admen(hrflow_profile: t.Dict) -> t.Dict:
    admen_profile = dict(ID_PERSONNE=int(hrflow_profile["reference"]), SOFT_DELETED=1)
    return admen_profile


DESCRIPTION = """
Need an efficient, high-performance HR solution for managing your applications?
Choose AD-Men, the No. 1 software for recruitment agencies.
"""

Admen = Connector(
    name="AD-MEN",
    type=ConnectorType.CRM,
    subtype="admen",
    description=DESCRIPTION,
    url="https://www.ad-rh.com/",
    warehouse=AdmenWarehouse,
    flows=(
        Flow(Mode.create, Entity.job, Direction.inbound, format=format_job),
        Flow(Mode.update, Entity.job, Direction.inbound, format=format_job),
        Flow(
            Mode.archive,
            Entity.job,
            Direction.inbound,
            format=format_job_archive_in_hrflow,
        ),
        Flow(
            Mode.create,
            Entity.profile,
            Direction.inbound,
            format=format_admen_profile_to_hrflow,
        ),
        Flow(
            Mode.update,
            Entity.profile,
            Direction.inbound,
            format=format_admen_profile_to_hrflow,
        ),
        Flow(
            Mode.archive,
            Entity.profile,
            Direction.inbound,
            format=format_profile_archive_in_hrflow,
        ),
        Flow(
            Mode.create,
            Entity.profile,
            Direction.outbound,
            format=format_hrflow_profile_to_admen,
        ),
        Flow(
            Mode.update,
            Entity.profile,
            Direction.outbound,
            format=format_hrflow_profile_for_update_to_admen,
        ),
        Flow(
            Mode.archive,
            Entity.profile,
            Direction.outbound,
            format=format_profile_archive_in_admen,
        ),
    ),
)



from hrflow_connectors.v2.connectors.francetravail.aisles import (
    AuthParameters,
    JobsAisle,
)
from hrflow_connectors.v2.core.warehouse import Warehouse

FranceTravailWarehouse = Warehouse(auth=AuthParameters, aisles=(JobsAisle,))



from hrflow_connectors.v2.connectors.adzuna.aisles import AuthParameters, JobsAisle
from hrflow_connectors.v2.core.warehouse import Warehouse

AdzunaWarehouse = Warehouse(auth=AuthParameters, aisles=(JobsAisle,))



from hrflow_connectors.v2.connectors.smartrecruiters.aisles import (
    AuthParameters,
    JobsAisle,
    ProfilesAisle,
)
from hrflow_connectors.v2.core.warehouse import Warehouse

SmartRecruitersWarehouse = Warehouse(
    auth=AuthParameters, aisles=(JobsAisle, ProfilesAisle)
)
